/* $Id: CalcPi.java,v 1.1 2009/03/10 21:15:29 rkufrin Exp $ */

/****************************************************************************
 *
 *                    University of Illinois/NCSA
 *                        Open Source License
 * 
 *          Copyright(C) 2004-2009, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                            Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * CalcPi: Java threaded version of approximation of pi.
 *         This example shows the use of the PerfSuite Java API.
 *
 *         Five lines of code are added to this class to obtain
 *         performance monitoring:
 *
 *         - import of the org.perfsuite.hwpc package.
 *         - declaration and initialization of PS_hwpcThreaded
 *             object through the getInstance() factory method
 *             Also, a constant string is declared for the output
 *             file name prefix.
 *         - monitoring is started within each thread
 *         - monitoring is stopped and written out before the
 *           end of each thread's run() method.
 *
 *         Search for the string "HWPC" to locate these lines.
 *
 * Author: Rick Kufrin
 * Date:   Feb. 13, 2004
 *
 * This file is part of PerfSuite.
 */

import java.util.*;
import org.perfsuite.hwpc.*;     // For HWPC classes

public class CalcPi {
    // Set up for HWPC monitoring
    static PS_hwpcThreaded hwpc = PS_hwpcThreaded.getInstance();
    static final String XMLPrefix = "CalcPi";

    static final String usage =
          "Usage: CalcPi numberOfIntervals numberOfThreads";
    static double pi = 0.0;
    static double w;
    static long intervals;
    static int nthreads;

    static class PiWorker extends Thread {
        private int myid;

        // Constructor stores each thread's assigned ID
        public PiWorker(int id) {
            myid = id;
            setName("PiWorker" + myid);
            start();
        }

        // Calculate this threads's contribution to overall approximation
        public void run() {
            double partialsum = 0.0;
            double x;

            System.out.println("hello from thread " + myid);

            // Start HWPC monitoring
            hwpc.start();

            for ( long i = myid+1; i <= intervals; i += nthreads ) {
                x = w * ((double) i - 0.5);       // 3 flops
                partialsum += 4.0 / (1.0 + x*x);  // 4 flops
            }

            sum(w * partialsum);

            // Stop HWPC monitoring and write data to XML
            hwpc.stop(XMLPrefix);
        }

        public synchronized void sum(double val) {
            pi += val;
        }
    }

    public static void main(String[] args) {
        // Check for proper arguments supplied
        if ( args.length != 2 ) {
           System.err.println(usage);
           return;
        }

        // Parse arguments and set global variables
        intervals = Long.parseLong(args[0]);
        nthreads = Integer.parseInt(args[1]);
        w = 1.0 / (double) intervals;
        String gt1 = "";
        if ( nthreads > 1 ) gt1 = "s";
        System.out.println("Run parameters: " + intervals + " intervals over " 
                           + nthreads + " thread" + gt1 + ".");
        System.out.println("Approx. # FP operations/thread: "
                           + Math.round((intervals / (double) nthreads) * 7.0));


        /*
         * Create the number of threads requested and wait for them
         * to complete their calculation
         */
        Thread workers[] = new Thread[nthreads]; 
        for ( int i = 0; i < nthreads; i++ ) {
            workers[i] = new PiWorker(i);
        }
        for ( int i = 0; i < nthreads; i++ ) {
            try {
                workers[i].join();
            }
            catch (InterruptedException e) {
                throw new RuntimeException(e);
            }
        }

        System.out.println(" Computed pi = " + pi);
        System.out.println("Java Math.PI = " + Math.PI);
        System.out.println("       Error = " + Math.abs(Math.PI - pi));
    }
}
