/* $Id: cpi-pthreads.c,v 1.14 2009/10/16 19:11:28 ruiliu Exp $ */

/*
 * cpi-pthreads: approximate the value of pi.
 *
 * Can be optionally compiled for POSIX threads with -D_REENTRANT 
 * and/or PerfSuite libpshwpc with -DPSHWPC.
 *
 * This program was adapted from an example appearing in the
 * article "OpenMP: A Proposed Industry Standard API for
 * Shared Memory Programming" by Leo Dagum (http://www.openmp.org/)
 *
 */

#if defined(_REENTRANT)
# include <pthread.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/syscall.h>

#if defined(PSHWPC)
# include <pshwpc.h>
#endif

#if defined(_REENTRANT)
static pthread_mutex_t reduction_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t tid_mutex       = PTHREAD_MUTEX_INITIALIZER;
static int id = 0;
static pthread_t *tid;
#endif

static int num_threads = 1;
static long long intervals;
static double pi, w;

double f(double a);

inline double
f(double a)
{
    return ( 4.0 / (1.0 + a*a) );
}

static void *
PIworker(void *arg)
{
    int myid = 0;  /* Initialize in case we are single-threaded */
    long long i;
    double sum, mypi, x;
#if defined(_REENTRANT)
    int ret;
#endif

#if defined(_REENTRANT)
    /* set individual id to start at 0 */
    ret = pthread_mutex_lock(&tid_mutex);
    if ( ret != 0 ) {
        perror("pthread_mutex_lock");
        pthread_exit(NULL);
    }
    myid = id++;
    ret = pthread_mutex_unlock(&tid_mutex);
    if ( ret != 0 ) {
        perror("pthread_mutex_unlock");
        pthread_exit(NULL);
    }
    printf("hello from thread %d [PID %d", myid, getpid());
# if defined(SYS_gettid)
    printf(", TID %ld", syscall(SYS_gettid));
# endif
    puts("]");
#endif

#if defined(PSHWPC) && defined(_REENTRANT)
    ret = ps_hwpc_start();
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "ps_hwpc_start: %s\n", ps_strerror(ret));
        pthread_exit(NULL);
    }
#endif

    /* integrate function */
    sum = 0.0;
    for (i = myid+1; i <= intervals; i += num_threads) {
        x = w * ((double) i - 0.5);
        sum += f(x);
    }

#if defined(PSHWPC) && defined(_REENTRANT)
    ret = ps_hwpc_stop("pi-worker");
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "ps_hwpc_stop: %s\n", ps_strerror(ret));
        pthread_exit(NULL);
    }
#endif

    mypi = w*sum;

    /* reduce value */
#if defined(_REENTRANT)
    ret = pthread_mutex_lock(&reduction_mutex);
    if ( ret != 0 ) {
        perror("pthread_mutex_lock");
        pthread_exit(NULL);
    }
#endif
    pi += mypi;
#if defined(_REENTRANT)
    ret = pthread_mutex_unlock(&reduction_mutex);
    if ( ret != 0 ) {
        perror("pthread_mutex_lock");
        pthread_exit(NULL);
    }
#endif

#if ! _REENTRANT || defined(NO_PTHREAD_EXIT)
    return((void *) 0);
#else
    pthread_exit(NULL);
#endif
}

int
main(int argc, char **argv)
{
    int argcexp;
#if defined(_REENTRANT)
    int i;
#endif
#if defined(PSHWPC)
    int ret;
#endif
  
#if defined(_REENTRANT)
    argcexp = 3;
#else
    argcexp = 2;
#endif

    /* check command line */
    if ( argc != argcexp ) {
        fprintf(stderr, "Usage: %s num-intervals", argv[0]);
#if defined(_REENTRANT)
        fputs(" num-threads", stderr);
#endif
        fputs("\n", stderr);
        exit(1);
    }

#if defined(PSHWPC)
    ret = ps_hwpc_init();
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "ps_hwpc_init: %s\n", ps_strerror(ret));
        if ( ret == PS_ENOTINSTALLED ) {
            /* Exit status of 77 is meaningful for autotest suites */
            exit(77);
        }
        exit(1);
    }
    ret = ps_hwpc_start();
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "ps_hwpc_start: %s\n", ps_strerror(ret));
        exit(1);
    }
#endif

    /* get num intervals and num threads from command line */
    sscanf(argv[1], "%lld", &intervals);
#if defined(_REENTRANT)
    num_threads = atoi(argv[2]);
#endif
    if ( intervals < 0 ) {
        fprintf(stderr, "Error: bad interval count (%lld). Exiting.\n",
                intervals);
        exit(1);
    }
    printf("Run parameters: %lld intervals over %d thread%s.\n",
           intervals, num_threads, (num_threads >1 ) ? "s": "");
    w = 1.0 / (double) intervals;
    pi = 0.0;

#if defined(_REENTRANT)
    tid = (pthread_t *) calloc(num_threads, sizeof(pthread_t));

    /* create the threads */
    for (i = 0; i < num_threads; i++) {
        if ( pthread_create(&tid[i], NULL, PIworker, NULL) ) {
            fprintf(stderr, "Cannot create thread %d\n", i);
            exit(1);
        }
    }

    /* join threads */
    for (i = 0; i < num_threads; i++) {
        pthread_join(tid[i], NULL);
    }
#else
    PIworker(NULL);
#endif

    printf("computed pi = %.16f\n", pi);

#if defined(PSHWPC)
    ret = ps_hwpc_stop("pi-main");
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "ps_hwpc_stop: %s\n", ps_strerror(ret));
        exit(1);
    }
    ret = ps_hwpc_shutdown();
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "ps_hwpc_shutdown: %s\n", ps_strerror(ret));
        exit(1);
    }
#endif

    exit(0);
}
