#!/usr/bin/perl

# $Id: psrunfull.pl,v 1.5 2006/01/21 22:45:36 rkufrin Exp $

############################################################################
#
#                      University of Illinois/NCSA
#                          Open Source License
#
#          Copyright(C) 2005-2006, The Board of Trustees of the
#              University of Illinois. All rights reserved.
#
#                             Developed by:
#
#                        The PerfSuite Project
#            National Center for Supercomputing Applications
#              University of Illinois at Urbana-Champaign
#
#                   http://perfsuite.ncsa.uiuc.edu/
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# + Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS WITH THE SOFTWARE.
############################################################################

#############################################################################
#
#  psrunfull - a "perfex -y -a" emulator for Linux using PerfSuite
#
#  This script compiles a list of all events available on the current
#  platform and then iterates across that set of events in groups of up to
#  32 (since only 32 can be active at once), running the program once for each
#  set of events, until all events have been used.  This data is then
#  combined and displayed in aggregate.  All XML files generated by this
#  script are cleaned up when it exits.
#
#  Author: Kalev Leetaru
#  Date:   6/29/2005
#
#############################################################################


# First make sure that we were given arguments, otherwise print help
# and exit...

if ($ARGV[0] eq "") {
    print "\nUsage: psrunfull [number_of_events] command\n";
    print "\nExamples:\n";
    print "\tpsrunfull myprogram   - runs myprogram and multiplexes across\n";
    print "\t                        all available events, 32 at a time\n";
    print "\tpsrunfull 1 myprogram - runs myprogram with all available\n";
    print "\t                        events, one event per run\n";
    exit;
}

# Generate a config file that will give us all the available events on
# this system...

# First compile a full list of all events on the system...

foreach $line (`psinv -p`) {
    if ($line=~/(PAPI.*?):/) {push(@counters, $1);}
}
$countercount = scalar(@counters);

if ($countercount < 2) {
    print "\nUnable to continue: It appears that the PerfSuite tools\n";
    print "(psinv, psrun) are not in your path!\n";
    exit;
}

# Check how many events the user wants us to use at once...

$countersatatime = 32;
if ($ARGV[0]=~/\d+/) {
    $countersatatime = $ARGV[0];
    $program = $ARGV[1];
    if ($#ARGV > 2) {
        for($i=2;$i<=$#ARGV;$i++) {$program .= " " . $ARGV[$i];}
    }
} else {
    $program = $ARGV[0];
    if ($#ARGV > 1) {
      	for($i=1;$i<=$#ARGV;$i++) {$program .= " " . $ARGV[$i];}
    }
}

print "RUNNING PROGRAM '$program'\n";

if ($countersatatime > 32) {
    print "\n\tWARNING: Multiplexing supports a maximum of 32 events at a time, resetting to 32 events.\n";
}
print "\tUsing $countersatatime events per run...\n";

# Let the user know how many events we found and warn if we are going to
# have to loop...

print "$countercount events found...\n";
if ($countercount > $countersatatime) {
    $execcount = int($countercount / $countersatatime) + 1;
    print "\n\tWARNING: $countercount events found (max $countersatatime per run)\n"; 
    print "Program will be executed $execcount times...\n\n";
}

# Enter a loop where we execute the program multiple times, $countersatatime
# counters-at-a-time...

for($i=0;$i<$countercount;$i+=$countersatatime) {

    # Generate our config file for this event batch...
    open(FILE, ">./all.xml");
    print FILE "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
    print FILE "<ps_hwpc_eventlist class=\"PAPI\" generator=\"psrunfull\">\n";
    for($j=$i;$j<($i+32);$j++) {
        if ($j < $countercount) {
            print FILE "<ps_hwpc_event type=\"preset\"";
            print FILE " name=\"$counters[$j]\" />\n";}
    }
    print FILE "</ps_hwpc_eventlist>\n";
    close(FILE);

    # Now run the given app via psrun with this config file...
    system("psrun -p -o \"lightweight\" -c ./all.xml $program");

} # End loop across all events...

# Now remove our config xml file...

unlink("./all.xml");

# ... and hand off all of the xml files to be combined...

system("psprocess -c ./lightweight*.xml > full.xml");

# Display output to the user...

system("psprocess full.xml");

# Lastly clean up all of the XML files we just generated...

foreach $file (glob("./lightweight*.xml")) {
    unlink($file);
}
unlink("full.xml");
