#!/bin/sh
# the next line restarts using tclsh \
exec tclsh "$0" "$@"

# $Id: hwpcfilter.tcl,v 1.28 2006/01/21 22:47:44 rkufrin Exp $

############################################################################
#
#                      University of Illinois/NCSA
#                          Open Source License
# 
#          Copyright(C) 2004-2006, The Board of Trustees of the
#              University of Illinois. All rights reserved.
#
#                             Developed by:             
# 
#                        The PerfSuite Project
#            National Center for Supercomputing Applications
#              University of Illinois at Urbana-Champaign
# 
#                   http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

# hwpcfilter.tcl - create a <hwpcreport> (counting mode) document from
#                  data filtered from one or more profiling runs.
#
# This is an example program that shows one way of filtering out
# samples gathered from a profiling report to learn about selected
# portions of an application.  This could be helpful if (for example)
# you wanted to know the cache hit rate for a loop that is in
# between lines 20 and 30 in the file "/home/me/prog/prog.c".
#
# Note that the range of lines specified need not be restricted
# only to a loop - you're free to specify any range of lines within
# a source file to examine the performance information relevant to
# those lines.  Or you can place no restrictions at all, in which
# case this program will "extrapolate" from one or more statistical 
# profiles to an approximation of the counting equivalents.
#
# This program requires no external software support beyond Tcl and 
# the tDOM Tcl extension.  The only change you may need to make is to
# set the correct location for the tDOM library installation on your
# system (see the "lappend auto_path" statement below)
#
# This program works by using tDOM and XPath to select certain nodes 
# from a PerfSuite <multihwpcprofilereport> document.  These documents
# can be generated by the following procedure:
#
# 1. use psrun to profile your application one or more times, selecting
#    a different hardware event each time.  You must use the same
#    host (computer) for each of these runs.
# 2. use psprocess to map the profiling samples to line numbers and write 
#    the results to a <hwpcprofilereport> document for each prior run
# 3. use psprocess again to create a combined <multihwpcprofilereport>
#    from the individual <hwpcprofilereport> documents
# 4. run this program on the result to create a <hwpcreport> document
#    that reflects only the hardware counts measured from the selected
#    portions of the program
# 5. use psprocess once more on the <hwpcreport> document to get the 
#    final result, as if it were a counting hwpcreport
#
# Example (assuming the program is in the file "prog"):
#
#   First, generate the profiles (this process can be repeated, and note
#   that the psprocess option "-x" is very important to include):
#
#   % psrun -c profile_cycles.xml -o cycles prog
#   % psprocess -e prog -x cycles.*.xml > cyclesXML
#   % psrun -c profile_l2tcm.xml -o l2tcm prog
#   % psprocess -e prog -x l2tcm.*.xml > l2tcmXML
#
#   The above sequence can easily be automated within a shell script.
#
#   Next, create a combined document containing the profiles:
#
#   % psprocess -c *XML > combined.xml
#
#   Filter out the lines from the file you want. Note that once you
#   have the combined file, you can do the next two steps as many times
#   as you want to look at different portions (often there are
#   only a few regions of interest in a large application).  Here, we
#   ask for estimated counts collected between lines 20 and 30 inclusive:
#
#   % hwpcfilter.tcl --file /home/me/prog/prog.c \
#                    --start 20 --end 30 combined.xml > counts.xml
#
#   Get the results, which have been deposited into a <hwpcreport>
#   document in "counting mode" format. psprocess automatically
#   computes and displays derived metrics based on the events measured
#   in addition to the raw event counts:
#
#   % psprocess counts.xml
#
# The functionality of this program may become a standard part of
# PerfSuite in the future, depending on its usefulness.  However,
# since there is a potential for a lot of inaccuracies and the procedure
# is subject to a lot of user modification, it currently remains an
# example program/utility.
#
# Author: Rick Kufrin
# Date:   2004/11/21
#
# This file is part of PerfSuite.
#

# Change this directory if needed to reflect the location of tDOM
# on your system

lappend auto_path /usr/local/lib

# Usage information procedure

proc usage {} {
    puts -nonewline "Usage: [file tail [info script]] "
    puts {[-f SOURCEFILE [[-s LINE] [-e LINE]]] XMLdoc}
}

# Procedure to parse the command line arguments

proc parseoptions { } {
    global argv filename startline endline XMLfile

    for {set optindex 0} {$optindex < [llength $argv]} {incr optindex} {
       set opt [lindex $argv $optindex]
       switch -glob -- $opt {
           -f -
           --f* {
               incr optindex
               set filename [lindex $argv $optindex]
           }
           -s -
           --s* {
               incr optindex
               set startline [lindex $argv $optindex]
           }
           -e -
           --e* {
               incr optindex
               set endline [lindex $argv $optindex]
           }
           -* {
               error "Unrecognized option: \"$opt\""
           }
           default {
               set XMLfile $opt
           }
       }
    }
 
    if { ! [info exists XMLfile] } {
        error "No XML document supplied"
    }

    if { ([info exists startline] || [info exists endline]) \
         && (! [info exists filename]) } {
            error "The -file option must be supplied with -start or\
                  -end options"
    }

    if { [info exists startline] && [info exists endline] } {
        if { $startline > $endline } {
            error "-start argument must be less than or equal\
                   to -end argument"
        }
    }
}


# This procedure creates an XPath expression that corresponds to
# the options supplied by the user.

proc createXPath {} {
    global filename startline endline

    set xpath "hwpcprofile/file"

    if { [info exists filename] } {
        append xpath "\[@name=\"$filename\"\]"
    }

    append xpath "/function/line"

    if { [info exists startline] } {
        append xpath "\[(@lineno >= $startline)"
        if { [info exists endline] } {
            append xpath "and (@lineno <= $endline)"
        }
        append xpath "\]"
    } elseif { [info exists endline] } {
        append xpath "\[(@lineno <= $endline)\]"
    }

    return $xpath
}

# This procedure extracts the event samples that conform to
# the constraints set by the user and passed in as an XPath query.
# It returns a list of four elements - the first is the class, the
# second is the counting domain, and the third and fourth elements
# are lists.  The third element are the event types (preset or native),
# the fourth are the event counts.  Both have the form
# {eventname value eventname value ...}
#
# The counts in the returned list are scaled according to the profiling
# threshold that's contained in the individual profile, so is an
# approximation to the true number of events that occurred over the run.
#
# This procedure also checks to ensure that the performance measurement
# class is acceptable/supported (currently PAPI only) and that the 
# hostname and command contained in each profile are identical.  This
# is just to try to help avoid totally off-the-wall usage.

proc geteventcounts {xmlroot xpath} {
    set supportedclasses [list PAPI]
    set eventsfound {}

    # Execute the XPath query over each profile, and gather
    # up the total samples contained within the resulting
    # node sets

    set profiles [$xmlroot selectNodes hwpcprofilereport]
    foreach profile $profiles {
        # Extract the host name
        set host [[$profile selectNodes executioninfo/hostname] text]
        if { ! [info exists hosts] } {
           set hosts $host
        } elseif { $host != $hosts } {
           error "Multiple hosts found in profiles"
        }

        # Extract the profiling details
        set eventelt [$profile selectNodes hwpcprofile/hwpcevent]
        if { $eventelt == "" } {
            error "No <hwpcevent> element found while scanning a profile"
        }

        # Grab the attributes
        catch {array unset att}
        set attlist [$eventelt attributes]
        foreach attname $attlist {
           set att($attname) [$eventelt getAttribute $attname]
        }

        # Check the profiling class
        if { [lsearch $supportedclasses $att(class)] == -1 } {
            error "Unsupported profiling class found: \"$att(class)\""
        }
        lappend classesfound $att(class)

        # Get the domain
        lappend domainsfound $att(domain)

        # Get the event name
        if { $att(type) == "preset" } {
            set eventname [$eventelt getAttribute name]
        } else {
            set eventname [$eventelt text]
        }
        lappend eventtypes $eventname $att(type)

        # Check to make sure this event hasn't been seen before
        if { [lsearch $eventsfound $eventname] != -1 } {
            error "Found multiple profiles measuring \"$eventname\""
        }
        lappend eventsfound $eventname

        # Get the sampling period
        set thresholdelt [$profile selectNodes hwpcprofile/period]
        if { $thresholdelt == "" } {
            error "No <period> element found while scanning a profile"
        }
        set threshold [$thresholdelt text]

        # Run the query
        set nodes [$profile selectNodes $xpath]

        # If the query returned any results, extract the sample counts
        set totalsamples 0.0
        if { $nodes == "" } {
           error "No samples matched the specification"
        } else {
           foreach node $nodes {
               set samples [$node text]
               set totalsamples [expr {(double($samples) * $threshold) \
                                       + $totalsamples}]
               #incr totalsamples [$node text]
               #puts -nonewline [$node asXML]
            }
            #puts "Total samples: $totalsamples"
        }
        lappend eventcounts $eventname [format %.0f $totalsamples]
    }

    # Double check to see that all classes and domains were identical
    if { [llength [lsort -unique $classesfound]] != 1 } {
        error "Must have exactly one event class represented"
    }
    if { [llength [lsort -unique $domainsfound]] != 1 } {
        error "Must have exactly one counting domain represented"
    }

    return [list [lindex $classesfound 0] [lindex $domainsfound 0] \
                 $eventtypes $eventcounts]
}

# This procedure uses tDOM to create the output XML document.
# It duplicates the machine information from the first profile found
# in the multi-report and creates new nodes for the event counts that 
# were gathered from the samples.  The end result is a document that
# can be post-processed by psprocess as a counting report.

proc writeXML {profileroot eventdata} {
    global filename startline endline

    # The elements we'll duplicate from the first profile.  The ordering
    # is significant in this routine (see handling of <annotation> element)

    set nodelist [list date hostname domainname username command wallclock \
                       cputime machineinfo]

    # Create the XML document
    set xmldoc [dom createDocument "hwpcreport"]
    $xmldoc documentElement xmlroot
    $xmlroot setAttribute \
        version "0.1" generator [file rootname [file tail [info script]]] \
        mode "count" xmlns:perfsuite "http://perfsuite.ncsa.uiuc.edu/"

    # Reuse nodes from the first profile
    foreach nodetype $nodelist {
        set nodes [$profileroot selectNodes \
                   hwpcprofilereport/executioninfo/$nodetype]

        # PerfSuite version 0.6.1 did not output a <command> element,
        # so guard against that.

        if { $nodes != "" } {
            set node [lindex $nodes 0]
            $xmlroot appendChild $node
        }

        if { $nodetype == "command" } {
            # Make a note of what we're doing in the <annotation> element,
            # which precedes the <wallclock> element in the hwpcreport DTD.
            if { ! [info exists filename] } {
                set annotation "No filters applied (complete)"
            } else {
                if { ! [info exists startline] } {
                    set startline 1
                }
                if { ! [info exists endline] } {
                    set endline end
                }
                set annotation [format {%s (lines %s-%s)} \
                                $filename $startline $endline]
            }
            set annotationnode [$xmldoc createElement annotation]
            set annotationtext [$xmldoc createTextNode $annotation]
            $annotationnode appendChild $annotationtext
            $xmlroot appendChild $annotationnode
        }
    }

    # Create new nodes for the event data we've gathered from samples
    set eventselt [$xmldoc createElement hwpceventdata]
    $eventselt setAttribute class  [lindex $eventdata 0] \
                            domain [lindex $eventdata 1] \
                            multiplexed "no"
    array set eventtypes  [lindex $eventdata 2]
    array set eventvalues [lindex $eventdata 3]
    foreach event [array names eventvalues] {
        set eventelt [$xmldoc createElement hwpcevent]
        $eventelt setAttribute type $eventtypes($event) \
                               name $event derived "no"
        $eventselt appendChild $eventelt
        set countelt [$xmldoc createTextNode $eventvalues($event)]
        $eventelt appendChild $countelt
    }

    $xmlroot appendChild $eventselt

    # All done... write the new document to standard output.
    puts "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\" ?>"
    puts -nonewline "[$xmldoc asXML -indent 4]"
}
 
###################################################################
#                  Main program starts here                       #
###################################################################

# First, check to see if we were executed correctly.  If not,
# we save the time required to load tDOM.

if { [catch {parseoptions} res] } {
    usage
    puts stderr "Error: $res"
    exit 1
}

# Also check to see we can read the XML file.  Again, don't bother
# with tDOM if not.

if { [catch {open $XMLfile "r"} fd] } {
    puts stderr "Error: $fd"
    exit 1
}

# Now go ahead and load tDOM into the interpreter

if { [catch {package require tdom} res] } {
    puts stderr "Error loading tDOM"
    exit 1
}

# Acquire the data

set data [read $fd]
close $fd

# Get the XPath query

if { [catch {createXPath} xpath] } {
    puts stderr "Error: $xpath"
    exit 1
}

# Parse the XML document and get the document element

set xmldoc [dom parse $data]
set xmlroot [$xmldoc documentElement]

# Sanity check to see that we were given the right kind of document

set doctype [$xmlroot nodeName]
if { $doctype != "multihwpcprofilereport" } {
    puts stderr "Error: This program only works with\
                 \"multihwpcprofilereport\" documents."
    puts stderr "       Your document is of type \"$doctype\"."
    exit 1
}

# Do the deed

if { [catch {geteventcounts $xmlroot $xpath} eventdata] } {
    puts stderr "Error: $eventdata"
    exit 1
}

writeXML $xmlroot $eventdata

exit 0
