#!/bin/sh
# the next line restarts using tclsh \
exec tclsh "$0" "$@"

############################################################################
#
#                      University of Illinois/NCSA
#                          Open Source License
# 
#            Copyright(C) 2006, The Board of Trustees of the
#              University of Illinois. All rights reserved.
#
#                             Developed by:             
# 
#                        The PerfSuite Project
#            National Center for Supercomputing Applications
#              University of Illinois at Urbana-Champaign
# 
#                   http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

# maxwall: adjust the value associate with the <wallclock> elements in 
#          a <multihwpcreport> to the maximum value of the original set.
#
#          This might be of interest when working with threaded programs
#          where you would like to view the work accomplished over the
#          entire lifetime of the program.  For example, consider a program
#          with four threads (including the main thread), with an execution
#          pattern like:
#
# Start      A            B                C                       D  End
# |--------------------------------------------------------------------->
# Thread 0   |            |
# Thread 1   |----------------------------->
#            |            |
# Thread 2   |----------------------------------------------------->
#                         |
# Thread 3                |---------------------------------------->
#
#          Thread 0 computes from Start to End, while Thread 1 computes
#          from point A to point C.  Thread 2 computes from point A to
#          point D.  Thread 3 computes from point B to point D.
#          When viewing the program execution as a whole (Start to End),
#          an alternate view of performance might want to consider
#          Threads 1, 2, and 3 in the context of the total runtime of
#          the program rather than just the portions during which the
#          threads were active and being monitored.
#
#          This example program attempts to correct for this by
#          adjusting the "run time" of all threads to the maximum run time
#          contained in the original XML document.
#
#          Note: due to differing clock speeds associated with each 
#          measurement, there may be some discrepancies in the wall clock
#          time reported by psprocess (especially if supplied with data
#          collected on a cluster).
#
# Usage: maxwall.tcl XML-document
#
# The XML document should be of type <multihwpcreport>.
# Output is sent to standard output.
#
# Author: Rick Kufrin
# Date:   May 9, 2006
#
# This file is part of PerfSuite.

# Adjust this path to the location of the tDOM library on your system.
lappend auto_path /usr/local/lib

package require tdom

if { [llength $argv] != 1 } {
    puts stderr "Usage: [file tail $argv0] multihwpcreport"
    exit 1
}

# Input the XML document

if { [catch {open [lindex $argv 0] "r"} fd] } {
    puts stderr "Error: $fd"
    exit 1
}

set data [read $fd]
close $fd

# Parse the document and check that it's the correct type

if { [catch {dom parse $data} xmldoc] } {
    puts stderr $xmldoc
    exit 1
}
set xmlroot [$xmldoc documentElement]

set doctype [$xmlroot nodeName]
if { $doctype != "multihwpcreport" } {
    puts stderr "Error: Input document must be a counting <multihwpcreport>."
    puts stderr "       This document is of type <$doctype>."
    exit 1
}

# Extract all the <wallclock> elements in the document

set wallelts [$xmlroot getElementsByTagName wallclock]

foreach elt $wallelts {
    lappend ticklist [$elt text]
}

# Determine the maximum wallclock and replace all with its value

set max [lindex [lsort -decreasing -real $ticklist] 0]

foreach elt $wallelts {
    set child [$elt firstChild]
    $child nodeValue $max
}

# Output the result document

puts {<?xml version="1.0" encoding="UTF-8" ?>}
puts -nonewline [$xmldoc asXML -indent 4]

exit 0
