#!/bin/sh
# the next line restarts using tclsh \
exec tclsh "$0" "$@"

# $Id: multihwpc-merge.tcl,v 1.7 2006/01/21 22:47:44 rkufrin Exp $

############################################################################
#
#                    University of Illinois/NCSA
#                        Open Source License
# 
#          Copyright(C) 2005-2006, The Board of Trustees of the
#              University of Illinois. All rights reserved.
#
#                            Developed by:             
# 
#                        The PerfSuite Project
#            National Center for Supercomputing Applications
#              University of Illinois at Urbana-Champaign
# 
#                   http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

# multihwpc-merge: generate a <hwpcreport> from the contents of
#                  a combined <multihwpcreport>.  The resulting
#                  document contains the machine and run information
#                  found in the first <hwpcreport> followed by the
#                  event counts extracted from each of the individual
#                  measurements.
#
#                  If no filename input argument is supplied, the
#                  data is read from standard input.
#                  The result is written to standard output.
#
#                  With the exception of the event counts, all other
#                  elements in the output document are the first
#                  occurrences as found in the input (<date>,
#                  <machineinfo>, and so on)
#
#                  If multiple occurrences of an identical event are
#                  found, only the first is retained and a message is
#                  written to standard error.
#
#                  This is a "poor man's" way of multiplexing and may
#                  be useful for generating derived metrics.  It is
#                  not included as a standard option for psprocess
#                  because of the potential for inaccuracy.
#
# Example: psprocess -c *.xml | multihwpc-merge.tcl
#
# See also the "runall.tcl" example, which can be used to run a command
# with all available events.
#
# Author: Rick Kufrin
# Date:   August, 2005
#
# This file is part of PerfSuite.
#

if { [llength $argv] > 1 } {
    puts stderr "Usage: $argv0 \[file.xml\]"
    exit 1
}

if { [llength $argv] == 0 } {
    set fd stdin
} else {
    if { [catch {open [lindex $argv 0] "r"} fd] } {
        puts stderr "Error: $fd"
        exit 1
    }
}

# Add the location of the tDOM library to Tcl's load path
lappend auto_path /usr/local/lib

if { [catch {package require tdom} res] } {
    puts stderr "Error loading tDOM Tcl extension"
    exit 1
}

# Read the original XML document

set rawdata [read $fd]

close $fd

# Use tDOM to parse the XML document and obtain the document root
# element.

if { [catch {dom parse $rawdata} domdocin] } {
    puts stderr "Error parsing input: $domdocin"
    exit 1
}
unset rawdata

set docrootin [$domdocin documentElement]
set doctype   [$docrootin nodeName]
if { $doctype != "multihwpcreport" } {
    puts stderr "Error: This program requires a <multhwpcreport> document."
    puts stderr "       Input was of type <$doctype>."
    puts stderr "Exiting."
    exit 1
}

# Create a new XML document for the output.

set docrootout [dom createDocument hwpcreport]

# Extract the first occurrence of initial elements and place
# in the output document.

$docrootout documentElement hwpcreport
$hwpcreport setAttribute \
    xmlns:perfsuite.ncsa.uiuc.edu http://perfsuite.ncsa.uiuc.edu/ \
    version 0.1 \
    generator multihwpc-merge \
    mode count

foreach element [list date hostname domainname username wallclock machineinfo] {
    set eltnode [$docrootin selectNodes hwpcreport/$element]
    if { $eltnode != "" } {
        $hwpcreport appendChild [lindex $eltnode 0]
    }
}

# Extract the overall event data container element.  This pulls out
# the first event data element, so note the names of events contained
# within for later use.

set data [lindex [$docrootin selectNodes hwpcreport/hwpceventdata] 0]
$hwpcreport appendChild $data
foreach event [$data childNodes] {
    lappend eventsseen [$event getAttribute name]
}

# Extract all remaining <hwpcevent> elements and place them in the
# output document.

set hwpcevents [$docrootin selectNodes hwpcreport/hwpceventdata/hwpcevent]

foreach event $hwpcevents {
    set eventname [$event getAttribute name]
    if { [lsearch $eventsseen $eventname] != -1 } {
        puts stderr "WARNING: multiple occurrences of \"$eventname\"\
                     (retaining only the first)"
        continue
    }
    lappend eventsseen $eventname
    $data appendChild $event
}

# Send result to standard output.

puts {<?xml version="1.0" encoding="UTF-8" ?>}
puts -nonewline [$docrootout asXML -indent 4]

exit 0
