#!/bin/sh
# the next line restarts using tclsh \
exec tclsh "$0" "$@"

############################################################################
#
#                      University of Illinois/NCSA
#                          Open Source License
# 
#          Copyright(C) 2005-2006, The Board of Trustees of the
#              University of Illinois. All rights reserved.
#
#                             Developed by:             
# 
#                        The PerfSuite Project
#            National Center for Supercomputing Applications
#              University of Illinois at Urbana-Champaign
# 
#                   http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

# $Id: runall.tcl,v 1.10 2006/01/21 22:47:44 rkufrin Exp $

# runall: run a given command multiple times, using every available
#         hardware performance event
#
# Usage: runall.tcl [-class classname] [-profile] [-help] command [args]
#
#         "classname": papi or perfmon
#
# Author: Rick Kufrin
# Date:   1/2/2005
#
# This file is part of PerfSuite.

# Change this to the location of your tDOM library installation

set TDOMLIBDIR /usr/local/lib

lappend auto_path $TDOMLIBDIR

# Some default variables that will be used later

set classatt       "PAPI"
array set classmap [list papi PAPI perfmon perfmon]
set skipevents     [list PAPI_FLOPS PAPI_IPS]
set configtag      "ps_hwpc_eventlist"
set eventtag       "ps_hwpc_event"

proc usage {{status 1}} {
    global argv0 classmap
    puts "Usage: [file tail $argv0] \[-class classname\] \[-profile\]\
                  \[-help\] command \[args\]"
    puts "       \"classname\" is one of: [array names classmap]"
    exit $status
}

# Generate a pshwpc configuration file with one or more events
# (this example only uses one event at a time)

proc makeconfig { filename class events } {
    global configtag eventtag

    set config [open $filename "w"]
    puts $config {<?xml version="1.0"?>}
    puts $config "<$configtag class=\"$class\">"
    foreach event $events {
        puts $config "\t<$eventtag type=\"preset\" name=\"$event\"/>"
    }
    puts $config "</$configtag>"
    close $config
}

# Check for options

while { [llength $argv] > 0 } {
    switch -glob -- [lindex $argv 0] {
        -class {
            set classatt [lindex $argv 1]
            set argv [lrange $argv 2 end]
        }
        -profile {
            set configtag "ps_hwpc_profile"
            set argv [lrange $argv 1 end]
        }
        -help {
            usage 0
        }
        -* {
            usage
        }
        * {
            break
        }
    }
}

# Make sure class name is in lowercase and check if it exists

set class [string tolower $classatt]
if { ! [info exists classmap($class)] } {
    puts stderr "Error: invalid class \"$classatt\""
    exit 1
}

# Were we given a command to run?

if { [llength $argv] < 1 } {
    puts "Error: no command was given"
    exit 1
}

# Load the performance counter class support

package require $class
namespace import ${class}::*

# Is the requested performance class installed on this computer?

set installed [eval set ${class}::installed]
if { ! $installed } {
    puts stderr "Error: $class does not appear to be installed on\
                 this computer."
    exit 1
}

# Run the command supplied with all available events on this computer

foreach event [availEvents] {
    # Avoid running with unsupported events
    if {[lsearch $skipevents $event] != -1 } {
        puts "Skipping event $event"
        continue
    }

    # Generate a configuration file for this run
    set config ${event}.xml
    if { [catch {makeconfig $config $classmap($class) $event} res] } {
        puts stderr "Error: $res"
        exit 1
    }

    # Let the user know what's going on
    puts -nonewline "Running \"$argv\" with $event... "
    flush stdout

    # Run the command
    set starttime [clock clicks -milliseconds]
    if { [catch {eval exec psrun -c $config $argv > /dev/null } res] } {
       puts stderr "\nError: failed to run with event \"$event\""
       puts stderr "Error: $res"
    } else {
       set endtime [clock clicks -milliseconds]
       set secs [expr {double($endtime - $starttime)*1.0e-3}]
       puts "done \[$secs secs\]"
    }

    # Get rid of the configuration file
    file delete $config
}

exit 0
