/* $Id: sampler.c,v 1.15 2006/09/14 13:39:40 rkufrin Exp $ */

/****************************************************************************
 *
 *                      University of Illinois/NCSA
 *                          Open Source License
 * 
 *            Copyright(C) 2006, The Board of Trustees of the
 *              University of Illinois. All rights reserved.
 * 
 *                             Developed by:             
 * 
 *                        The PerfSuite Project
 *            National Center for Supercomputing Applications
 *              University of Illinois at Urbana-Champaign
 * 
 *                   http://perfsuite.ncsa.uiuc.edu/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * + Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimers.
 * + Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimers in
 *   the documentation and/or other materials provided with the distribution.
 * + Neither the names of The PerfSuite Project, NCSA/University of Illinois
 *   at Urbana-Champaign, nor the names of its contributors may be used to
 *   endorse or promote products derived from this Software without specific
 *   prior written permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 ****************************************************************************/

/*
 * sampler: example program that uses interval timers to sample the
 *          values of performance monitoring periodically.
 *
 * Author: Rick Kufrin
 * Date:   1/29/2006
 *
 * This file is part of PerfSuite.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <pshwpc.h>

#define DEFAULT_PERIOD 5
#define DEFAULT_SAMPLES 10
#define DEFAULT_SIGNAL SIGVTALRM
#define DEFAULT_ITIMER ITIMER_VIRTUAL

volatile int samples_remaining = DEFAULT_SAMPLES;
int seconds = DEFAULT_PERIOD;
int write_results = 0;

struct itimerval timer;

int numevents = 0;
char **eventnames = NULL;
ps_hwpc_value_t *last, *current;

double a = 0.01;  /* Used in busyloop */

void busyloop(void);
void process_options(int argc, char **argv);
void sampler(int signo);
void setup_alarm(int seconds);
void setup_pshwpc(void);
void shutdown_pshwpc(int write_results);
void usage(void);


/*
 * main program parses user options, initializes and starts monitoring
 * through libpshwpc, installs an interval timer and handler.  It then
 * enters a busyloop that does simple floating point operations.  busyloop
 * returns when the requested number of samples have been taken and then
 * use of libpshwpc is terminated before the program exits.
 */

int main(int argc, char **argv)
{
    process_options(argc, argv);

    setup_pshwpc();

    setup_alarm(seconds);

    busyloop();

    shutdown_pshwpc(write_results);

    exit(0);
}


/*
 * usage prompt
 */

void usage(void)
{
    puts("Usage: sampler [-n samples] [-s seconds] -w");
    exit(0);
}


/*
 * parse the user-supplied command-line options.
 */

void process_options(int argc, char **argv)
{
    int arg;

    for (arg = 1; arg < argc; arg++ ) {
        if (strcmp(argv[arg], "-n") == 0 ) {
            arg++;
            if (arg == argc) {
               fputs("Option \"-n\" requires an argument.\n", stderr);
               exit(1);
            }
            samples_remaining = atoi(argv[arg]);
        }
        else if (strcmp(argv[arg], "-s") == 0 ) {
            arg++;
            if (arg == argc) {
               fputs("Option \"-s\" requires an argument.\n", stderr);
               exit(1);
            }
            seconds = atoi(argv[arg]);
        }
        else if (strcmp(argv[arg], "-w") == 0 ) {
            write_results = 1;
        }
        else {
            usage();
        }
    }

    printf("Generating %d samples at %d second intervals.\n",
           samples_remaining, seconds);
    printf("XML output will %sbe written.\n", write_results ? "" : "not ");
}


/*
 * busyloop is where all the time is spent in this program.  It
 * repeatedly does useless floating point operations until signaled
 * that the sampling is over via the variable samples_remaining, which
 * is decremented by the interrupt handler.
 */

void busyloop(void)
{
    do {
       a *= 1.0015;
    } while ( samples_remaining );

    return;
}


/*
 * setup_pshwpc initializes libpshwpc and allocates two buffers to hold
 * counter values.  The buffers will be switched at every interrupt.
 */

void setup_pshwpc(void)
{
    int ret;

    ret = ps_hwpc_init();
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "Initialization error: %s\n", ps_strerror(ret));
        exit(1);
    }

    ret = ps_hwpc_eventnames(&eventnames);
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "Error retrieving event names: %s\n", ps_strerror(ret));
        exit(1);
    }

    ret = ps_hwpc_numevents(&numevents);
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "Error retrieving event count: %s\n", ps_strerror(ret));
        exit(1);
    }

    last    = (ps_hwpc_value_t *) calloc(numevents, sizeof(ps_hwpc_value_t));
    current = (ps_hwpc_value_t *) calloc(numevents, sizeof(ps_hwpc_value_t));

    ret = ps_hwpc_start();
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "Error starting counting: %s\n", ps_strerror(ret));
        exit(1);
    }

    return;
}


/*
 * shutdown_pshwpc terminates use of libpshwpc and optionally writes
 * the final data out in an XML document by calling ps_hwpc_stop.
 */

void shutdown_pshwpc(int write_results)
{
    int ret;

    if ( write_results ) {
        ret = ps_hwpc_stop(NULL);
        if ( ret != PS_SUCCESS ) {
            fprintf(stderr, "Error writing results: %s\n", ps_strerror(ret));
        }
    }

    ret = ps_hwpc_shutdown();
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "Error shutting down: %s\n", ps_strerror(ret));
    }

    return;
}


/*
 * sampler is the "interesting" part of this program.  It is the interrupt
 * handler that reads the current values of the counters and decrements the
 * global variable that signals termination.  It prints out the difference
 * between the prior reading of the counters and the current reading, then
 * swaps the event count buffers for use upon next interrupt.
 */

void sampler(int signo)
{
    ps_hwpc_value_t *tmp;
    int i, ret;

    ret = ps_hwpc_read(current);
    if ( ret != PS_SUCCESS ) {
        fprintf(stderr, "Error reading values: %s\n", ps_strerror(ret));
        exit(1);
    }

    samples_remaining--;

    for ( i = 0; i < numevents; i++ ) {
        printf("%-18s: %lld\n", eventnames[i], current[i] - last[i]);
    }
    printf("%-18s: %d\n\n", "Samples remaining", samples_remaining);

    tmp = last;
    last = current;
    current = tmp;
}


/*
 * setup_alarm arranges for periodic interrupts to be sent to the
 * process every "seconds" seconds.  The "sampler" function is
 * designated as the interrupt handler.
 */

void setup_alarm(int seconds)
{
    struct sigaction act;
    int ret;

    memset(&act, 0, sizeof(struct sigaction));
    act.sa_handler = sampler;

    ret = sigaction(DEFAULT_SIGNAL, &act,  NULL);
    if ( ret != 0 ) {
        perror("Error installing signal handler");
        exit(1);
    }
       
    timer.it_interval.tv_sec = timer.it_value.tv_sec = (long) seconds;
    timer.it_value.tv_usec = timer.it_interval.tv_usec = 0L;

    ret = setitimer(DEFAULT_ITIMER, &timer, NULL);
    if ( ret != 0 ) {
        perror("Error setting interval timer");
        exit(1);
    }

    return;
}
