# $Id: combine.tcl,v 1.19 2009/01/09 21:38:00 rkufrin Exp $

############################################################################
#
#                    University of Illinois/NCSA
#                        Open Source License
# 
#        Copyright(C) 2003-2006, The Board of Trustees of the
#            University of Illinois. All rights reserved.
#
#                           Developed by:             
# 
#                      The PerfSuite Project
#           National Center for Supercomputing Applications
#             University of Illinois at Urbana-Champaign
# 
#                  http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

############################################################################
# Process all the files specified as the argument list
# Parse and check that they are of the same document type
# If successful, return a list of DOM nodes for further
# processing.
#
# Author: Rick Kufrin
# Date:   1/2003
#
# This file is part of PerfSuite

proc readAndValidateMultiple {fileList ignoreErrors} {
    global timers

    set res {}
    set doctype ""
    if { ! [info exists timers(total,domparse)] } {
        set timers(total,domparse) 0
    }

    foreach file $fileList {
        if { [catch {open $file "r"} fd] } {
            error $fd
        }
        set xmldata [read $fd]
        close $fd

        set starttime [clock clicks -milliseconds]
        if { [catch {dom parse $xmldata} xmldoc] } {
            if { $ignoreErrors } {
                continue
            } else {
                error "$xmldoc (while parsing \"$file\")"
            }
        }
        set stoptime [clock clicks -milliseconds]
        incr timers(total,domparse) [expr {$stoptime - $starttime}]

        set docRoot [$xmldoc documentElement]
        set type [$docRoot nodeName]

        if { $doctype == "" } {
            set doctype $type
            #puts "Type set to $type"
        } else {
            if { ! [string match $type $doctype]} {
                error "document type error: \"$type\" and \"$doctype\" found"
            }
        }

        lappend res $xmldoc

        #puts "processed $file, added $xmldoc to result"
    }

    return $res
}

proc doCombine { combineFileList {outputfd stdout} {ignoreErrors 1} } {
    global psprocessVersion options env

    # @@ Do this now or wait for error when actually processed?
    foreach file $combineFileList {
        if { [isuri $file] } {
            error "HTTP access not supported for combine mode"
        }
        if { ! [file exists $file] } {
            error "$file does not exist"
        }
        if {! [file readable $file] } {
            error "$file is not readable"
        }
    }

    if { [catch {readAndValidateMultiple $combineFileList $ignoreErrors} \
          docList] } {
        error $docList
    }

    if { [llength $docList] == 0 } {
        error "no valid input XML documents found"
    }

    # Special case treatment for vmon output.  Instead of writing
    # a <multihwpcprofile> document from the list of input XML
    # documents the user supplied, write a multi-profile vmon.out
    # format file

    if { $options(vmonout) } {
        source [file join [file dir [info script]] vmon.tcl]
        foreach doc $docList {
            set root [$doc documentElement]
            set profile [$root getElementsByTagName hwpcprofiledata]
            lappend profiles $profile
        } 
        vmon_write $outputfd $profiles
        return
    }

    # Build up the DOM tree and write it out.
    # If we're combining "multihwpcreport"s, then we don't want to
    # follow the convention of an additional "multi" prefix, instead
    # we'll make it a multi-run summary with a root element of
    # "hwpcreportcollection"
    # @@ Should validate the accepted types

    set documenttype [[[lindex $docList 0] documentElement] nodeName]
    if { [string match $documenttype "multihwpcreport"] } {
        set basetype "hwpcreportcollection"
    } else {
        set basetype "multi"
        append basetype $documenttype
    }

    set combinedDoc [dom createDocument $basetype]
    if { [info exists env(PS_DTDDIR)] } {
        $combinedDoc systemId "$env(PS_DTDDIR)/pshwpc/$basetype.dtd"
    }
    set combinedRoot [$combinedDoc documentElement]
    $combinedRoot setAttribute \
            xmlns:perfsuite "http://perfsuite.ncsa.uiuc.edu/" \
            version $psprocessVersion generator "psprocess"

    if { [info exists options(annotationString)] } {
        set annotationObj [$combinedDoc createElement annotation]
        $combinedDoc createTextNode $options(annotationString) annotationTextObj
        $annotationObj appendChild $annotationTextObj
        $combinedRoot appendChild $annotationObj
    }

    foreach doc $docList {
        set root [$doc documentElement]
        $combinedRoot appendChild $root
    }

    # set outputfd [open $outputfile "w"]

    puts $outputfd "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>"
    if { [$combinedDoc systemId] != "" } {
        puts $outputfd "[$combinedDoc asXML -indent 2 -doctypeDeclaration true]"
    } else {
        puts $outputfd "[$combinedDoc asXML -indent 2]"
    }
    close $outputfd

    # @@ What needs to be explicitly deleted? I had problems
    # when mixing obj/set styles
    foreach doc $docList {
        $doc delete
    }
    unset docList combinedDoc
    # @@ This crashed Tcl/Wish on Windows
    #    $combinedDoc delete

    return
}
