# $Id: cube.tcl,v 1.10 2009/09/30 21:09:28 rkufrin Exp $
############################################################################
#
#                      University of Illinois/NCSA
#                          Open Source License
# 
#          Copyright(C) 2008-2009, The Board of Trustees of the
#              University of Illinois. All rights reserved.
#
#                             Developed by:             
# 
#                        The PerfSuite Project
#            National Center for Supercomputing Applications
#              University of Illinois at Urbana-Champaign
# 
#                   http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################


# From profile.tcl: description of elements contained in the addrmap
# array (indices are PC values):
#
#           { samples module function file line }
#
# If information about the function and file is not available,
# their slots will contain "??".  Unknown line numbers will
# be reported as "0".
#
# Author: Rick Kufrin
# Date:   December 9, 2008
#
# This file is part of PerfSuite.

if { [catch {package require pscube} res ] } {
    error "could not load Cube support"
}

proc cube_write {cube proftree profileDoc} {
    global options
    set debug 0

    set pinfo [$profileDoc selectNodes hwpcprofiledata]
    set class [$pinfo getAttribute class] 

    set event [$profileDoc getElementsByTagName profileevent]
    set met   [$event getAttribute name]

    set hostname [$profileDoc getElementsByTagName hostname]
    set nod [$hostname text]
    set mach $nod

    # Use care with process and thread IDs, as they were not present
    # in PS XML files prior to version 1.0

    set procnode [$profileDoc getElementsByTagName pid]
    if { $procnode != "" } {
        set proc [$procnode text] 
    } else {
        set proc 0
    }
    set thrnode [$profileDoc getElementsByTagName thread]
    if { $thrnode != "" } {
        set thr [$thrnode text]
    } else {
        set thr 0
    }

    set metric [::cube::create $cube metric $met $met]

    set machine [::cube::create $cube machine $mach]
    set node [::cube::create $cube node $nod $machine]

    set process [::cube::create $cube process $proc $proc $node]
    set thread [::cube::create $cube thread $thr $thr $process]

    foreach progelt $proftree {
        set progname [lindex $progelt 1]
        set program [::cube::create $cube program $progname]

        foreach modelt [lsort -integer -decreasing -index 2 \
                           [lindex $progelt 3]] {
            set modulename [lindex $modelt 1]
            set module [::cube::create $cube module $modulename $program]

            foreach fileelt [lsort -integer -decreasing -index 2 \
                                [lindex $modelt 3]] {
                set filename [lindex $fileelt 1]
                set file [::cube::create $cube file $filename $module]

                foreach functionelt [lsort -integer -decreasing -index 2 \
                                        [lindex $fileelt 3]] {
                    set functionname [lindex $functionelt 1]
                    set function [::cube::create $cube function $functionname $file]

                    foreach lineelt [lsort -integer -decreasing -index 2 \
                                        [lindex $functionelt 3]] {
                        set lineno [lindex $lineelt 1]
                        set line [::cube::create $cube line $lineno $function]

                        set samples [lindex $lineelt 2]
                        ::cube::assign $cube $samples $metric $thread $line
                    }
                }
            }
        }
    }

    return
}

proc cube_append {cube proftree profileDoc} {
    global cube_cache
    set debug 0

    set event [$profileDoc getElementsByTagName profileevent]
    set met [$event getAttribute name]

    set hostname [$profileDoc getElementsByTagName hostname]
    set nod [$hostname text]
    set mach $nod

    set procnode [$profileDoc getElementsByTagName pid]
    if { $procnode != "" } {
        set proc [$procnode text]
    } else {
        set proc 0
    }
    set thrnode [$profileDoc getElementsByTagName thread]
    if { $thrnode != "" } {
        set thr [$thrnode text]
    } else {
        set thr 0
    }

    if { ! [info exists cube_cache(metric,$met)] } {
        set metric [::cube::create $cube metric $met $met]
        psdebug 2 "NO CACHED METRIC $met"
        set cube_cache(metric,$met) $metric
    } else {
        set metric $cube_cache(metric,$met)
        psdebug 2 "FOUND CACHED METRIC $met"
    }

    if { ! [info exists cube_cache(machine,$mach)] } {
        set machine [::cube::create $cube machine $mach]
        psdebug 2 "NO CACHED MACHINE $mach"
        set cube_cache(machine,$mach) $machine
    } else {
        set machine $cube_cache(machine,$mach)
        psdebug 2 "FOUND CACHED MACHINE $mach"
    }

    if { ! [info exists cube_cache(node,$nod,$mach)] } {
        set node [::cube::create $cube node $nod $machine]
        psdebug 2 "NO CACHED NODE $nod,$mach"
        set cube_cache(node,$nod,$mach) $node
    } else {
        set node $cube_cache(node,$nod,$mach)
        psdebug 2 "FOUND CACHED NODE $nod,$mach"
    }

    if { ! [info exists cube_cache(process,$proc,$nod)] } {
        set process [::cube::create $cube process $proc $proc $node]
        psdebug 2 "NO CACHED PROCESS $proc,$nod"
        set cube_cache(process,$proc,$nod) $process
    } else {
        set process $cube_cache(process,$proc,$nod)
        psdebug 2 "FOUND CACHED PROCESS $proc,$nod"
    }

    if { ! [info exists cube_cache(thread,$thr,$proc)] } {
        set thread [::cube::create $cube thread $thr $thr $process]
        psdebug 2 "NO CACHED THREAD $thr,$proc"
        set cube_cache(thread,$thr,$proc) $thread
    } else {
        set thread $cube_cache(thread,$thr,$proc)
        psdebug 2 "FOUND CACHED THREAD $thr,$proc"
    }

    foreach progelt $proftree {
        set progname [lindex $progelt 1]
        if { ! [info exists cube_cache(program,$progname)] } {
            set program [::cube::create $cube program $progname]
            psdebug 2 "NO CACHED PROGRAM $progname"
            set cube_cache(program,$progname) $program
        } else {
            set program $cube_cache(program,$progname)
            psdebug 2 "FOUND CACHED PROGRAM $progname ($program)"
        }

        foreach modelt [lsort -integer -decreasing -index 2 \
                           [lindex $progelt 3]] {
            set modulename [lindex $modelt 1]
            if { ! [info exists cube_cache(module,$modulename,$progname)] } {
                set module [::cube::create $cube module $modulename $program]
                psdebug 2 "NO CACHED MODULE $modulename,$progname"
                set cube_cache(module,$modulename,$progname) $module
            } else {
                set module $cube_cache(module,$modulename,$progname)
                psdebug 2 "FOUND CACHED MODULE $modulename,$progname ($module)"
            }

            foreach fileelt [lsort -integer -decreasing -index 2 \
                                [lindex $modelt 3]] {
                set filename [lindex $fileelt 1]
                if { ! [info exists cube_cache(file,$filename,$modulename)] } {
                    set file [::cube::create $cube file $filename $module]
                    psdebug 2 "NO CACHED FILE $filename,$modulename"
                    set cube_cache(file,$filename,$modulename) $file
                } else {
                    set file $cube_cache(file,$filename,$modulename)
                    psdebug 2 "FOUND CACHED FILE $filename,$modulename ($file)"
                }

                foreach functionelt [lsort -integer -decreasing -index 2 \
                                        [lindex $fileelt 3]] {
                    set functionname [lindex $functionelt 1]
                    if { ! [info exists \
                               cube_cache(function,$functionname,$filename)] } {
                        set function [::cube::create $cube function \
                                         $functionname $file]
                        psdebug 2 "NO CACHED FUNCTION $functionname,$filename"
                        set cube_cache(function,$functionname,$filename) \
                            $function
                    } else {
                        set function \
                            $cube_cache(function,$functionname,$filename)
                        psdebug 2 "FOUND CACHED FUNCTION $functionname,$filename ($function)"
                        }
                    }

                   foreach lineelt [lsort -integer -decreasing -index 2 \
                                       [lindex $functionelt 3]] {
                       set lineno [lindex $lineelt 1]
                       set samples [lindex $lineelt 2]

                       if { ! [info exists \
                                  cube_cache(line,$lineno,$functionname)] } {
                           set line \
                               [::cube::create $cube line $lineno $function]
                           psdebug 2 "NO CACHED LINE $lineno,$functionname"
                           set cube_cache(line,$lineno,$functionname) $line
                       } else {
                           set line $cube_cache(line,$lineno,$functionname)
                           psdebug 2 "FOUND CACHED LINE $lineno,$functionname ($line)"
                       }

                       ::cube::assign $cube $samples $metric $thread $line
                       psdebug 2 "EXE: ::cube::assign $cube $samples $metric $thread $line"
                   }
                }
            }
        }
    }

    return
}
