# $Id: filelist.tcl,v 1.5 2009/10/08 21:05:03 rkufrin Exp $

############################################################################
#
#                      University of Illinois/NCSA
#                          Open Source License
# 
#            Copyright(C) 2009, The Board of Trustees of the
#              University of Illinois. All rights reserved.
#
#                             Developed by:             
# 
#                        The PerfSuite Project
#            National Center for Supercomputing Applications
#              University of Illinois at Urbana-Champaign
# 
#                   http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

# processFileList: dispatch processing of files passed in via argument
#
# Author: Rick Kufrin
# Date:   1/15/2009
#
# This file is part of PerfSuite.
#

#
# This proc is work-in-progress, and is being incrementally added to psprocess.
#
# At present, the following are supported or expected to be added:
#
# 1. Multiple input files of type hwpcreport (mode=profile), using the option
#    -x.  The output in this case is an multihwpcprofilereport XML document,
#    with  PC->source mapping done.
#
# 2. Multiple input files of type hwpcreport (mode=profile), with the option
#    --cube requested.  The output in this case is a Cube-formatted parallel
#    profile. To be added.
#
# Also note, as is documented in psprocess.tcl, this proc is currently only
# reached when the option "--glob" has been used in psprocess invocation.

proc processFileList {fileList} {
    global env options timers

    set doctype ""
    set timers(total,domparse) 0

    foreach file $fileList {
        if { [catch {readxml $file} data] } {
            error "$data"
        }

        set timers(start,domparse) [clock clicks -milliseconds]
        if { $options(fast) } {
            if { [catch {dom parse -simple $data doc} res] } {
                error "XML parse (fast) of file \"$file\" failed ($res)"
            }
        } else {
            if { [catch {dom parse $data} doc] } {
                error "XML parse of file \"$file\" failed ($doc)"
            }
        }
        set timers(stop,domparse) [clock clicks -milliseconds]
        set parsetime [expr {$timers(stop,domparse) - $timers(start,domparse)}]
        psdebug 3 "Parse time for $file: [expr {$parsetime / 1000.0}] seconds"\
            stderr
        incr timers(total,domparse) $parsetime

        set root [$doc documentElement]
        set type [$root nodeName]
        if { $doctype == "" } {
            set doctype $type
        } elseif { $type != $doctype } {
            error "Non-uniform document types found (\"$doctype\",\"$type\")\
                   when processing input files."
        }

        switch $doctype {
            hwpcreport {
                set mode [$root getAttribute mode]
                if {$mode != "profile"} {
                    error "multiple input documents of this type require -c\
                           option"
                } elseif { [llength \
                                [info procs \
                                     hwpcreportCreateProfileReportElement]] \
                               == 0 } {
                    source [file join [file dirname [info script]] \
                                profilexml.tcl]
                }


                if { $options(xmlout) } {
                    if { ! [info exists newdoc] } {
                        set newdoctype "multihwpcprofilereport"
                        set newdoc [dom createDocument $newdoctype]
                        set version 0.2

                        # DTD
                        if { [info exists env(PS_DTDDIR)] } {
                            $newdoc systemId \
                                "$env(PS_DTDDIR)/pshwpc/$newdoctype-$version.dtd"
                        }

                        set newroot [$newdoc documentElement]
                        $newroot setAttribute \
                            version "0.2" \
                            generator "psprocess" \
                            xmlns:perfsuite "http://perfsuite.ncsa.uiuc.edu/"
                    }
                    hwpcreportCreateProfileReportElement $newroot $root
#                    $newroot appendChild \
#                        [hwpcreportCreateProfileReportElement $newdoc $root]
                    
                } elseif { $options(cube) } {
                    if { ! [info exists cube] } {
                        foreach f { cube profile } {
                            source [file join [file dirname [info script]] \
                                $f.tcl]
                        }
                        if { ! [info exists options(executable)] } {
                            set options(executable) ""
                        }
                        set cube [::cube::new]
                    }
                    set profiledata [$root selectNodes hwpcprofiledata]
                    set proftree [createmappedprofile $profiledata \
                                  $options(executable) $options(showprogress)]
                    cube_append $cube $proftree $root
                } else {
                    error "processFileList: currently requires -x or --cube\
                           option"
                }
            }

            multihwpcreport -
            hwpcprofilereport -
            multihwpcprofilereport {
                    error "processFileList: support for \"$doctype\"\
                           not yet implemented"
            }

            default {
                error "processFileList: support for \"$doctype\"\
                       not yet implemented"
            }
        }

#        $doc delete
    }

    if { $options(xmlout) } {
        if [ info exists options(outputFile) ] {
            if { [catch {open $options(outputFile) "w"} fd] } {
                error $fd
            }
        } else {
            set fd stdout
        }

        puts $fd {<?xml version="1.0" encoding="UTF-8" standalone="no" ?>}

        # Only attempt to output the DTD in the result if it was set
        # in the hwpcreportCreateProfileReportXML procedure.
        if { [$newdoc systemId] != "" } {
            puts $fd "[$newdoc asXML -indent 2 -doctypeDeclaration true]"
        } else {
            puts $fd "[$newdoc asXML -indent 2]"
        }
    } elseif { $options(cube) } {
        if { [info exists options(outputFile)] } {
            set fd $options(outputFile)
        } else {
            set fd stdout
        }
        ::cube::dump $cube $fd
        ::cube::delete $cube
    }
}
