# $Id: gmon.tcl,v 1.12 2009/11/24 16:43:31 rkufrin Exp $

############################################################################
#
#                    University of Illinois/NCSA
#                        Open Source License
# 
#          Copyright(C) 2004-2006, The Board of Trustees of the
#              University of Illinois. All rights reserved.
#
#                            Developed by:             
# 
#                        The PerfSuite Project
#            National Center for Supercomputing Applications
#              University of Illinois at Urbana-Champaign
# 
#                   http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

# gmon.tcl - convert libpshwpc profiling data to gprof gmon format
#
# Author: Rick Kufrin
# Date:   May 2004
#
# This file is part of PerfSuite.

# @@ Needs lots of work - currently only suitable for 32-bit system
# The scaling of bins might not be matching perfectly with gprof,
# check with -d option for [hist_read_rec] output
# @@ Masks may not be necessary

#struct gmon_hdr
#  {
#    char cookie[4];
#    char version[4];
#    char spare[3 * 4];
#  };

proc gmon_preamble {fd} {
    puts -nonewline $fd [binary format a4 "gmon"];             # GMON_MAGIC
    puts -nonewline $fd [binary format i [expr {1 & 0xFFFF}]]; # GMON_VERSION
    for { set i 0 } { $i < 3 } { incr i } {
        puts -nonewline $fd [binary format i 0x00000000];      # padding
    }
}

# Write a gmon-format histogram.
# Format is:
#  Header:
#    1 byte : gmon.out "tag" (only histograms done here, so 0 from gmon_out.h)
#  4/8 bytes: low PC
#  4/8 bytes: high PC
#    4 bytes: number of bins
#    4 bytes: hertz (rate of profiling clock)
#   15 bytes: long description ("i-cache misses")
#    1 byte : short description ("1")
# = 33 bytes on 32-bit, 41 on 64-bit
#
#  Histogram:
#  Array of 16-bit histogram values
#
# Refs: hist.c (function hist_write_hist), gmon_out.h, gmon_io.c
#      in binutils/gprof,  gprof info page (Profiling Details)

proc gmon_writeprofile {fd profile} {
    global tcl_platform tcl_version
    set DEBUG 0

    # tag
    set GMON_TAG_TIME_HIST 0
    puts -nonewline $fd [binary format c [expr {$GMON_TAG_TIME_HIST & 0xFF}]]

    set profiles [$profile getElementsByTagName module]
    set mainprofile [lindex $profiles 0]
    set mainsamples [$mainprofile getElementsByTagName sample]

    if { $mainsamples == "" } {
        set profileeventelt [$profile getElementsByTagName profileevent]
        set desc      [$profileeventelt getAttribute name]
        error "No samples recorded in this profile (measurement: $desc)"
    }
    set lowelt  [lindex $mainsamples 0]
    set highelt [lindex $mainsamples end]

    set lowpcstr "[$lowelt getAttribute pc]"
    set lowpc    "0x$lowpcstr"
    set highpc  "0x[$highelt getAttribute pc]"
    if { [expr {(($highpc - $lowpc) / 2) % 2}] != 0 } {
        set highpc "0x[format {%x} [expr {$highpc + 2}]]"
    }
    if { $DEBUG } {
        puts "lowpc = $lowpc highpc = $highpc"
    }

    # Tcl 8.4+ supports the "w" format to output 64-bit integers
    # in little-endian format.  Could do this "by hand", but 8.4
    # has been around long enough to be the standard now.
    set systems64bit [list ia64 x86_64]
    if { [lsearch $systems64bit $tcl_platform(machine)] != -1 } {
        if { $tcl_version < 8.4 } {
            error "gmon output for this platform ($tcl_platform(machine))\
                   requires Tcl version 8.4 or higher"
        } else {
            set fmt w
            set mask 0xffffffffffffffff
        }
    } else {
        set fmt i
        set mask 0xffffffff
    }
    unset systems64bit

    # @@ Not sure if mask is necessary for these
    # low PC
    puts -nonewline $fd [binary format $fmt [expr {$lowpc & $mask}]]

    # high PC, but first increment by 2 to ensure there are sufficient
    # bins to accommodate the samples that occurred
    incr highpc 2
    puts -nonewline $fd [binary format $fmt [expr {$highpc & $mask}]]

    # number of bins
    set nbins [expr {($highpc - $lowpc) / 2}]
    if { $DEBUG } {
        puts "nbins = $nbins"
    }
    puts -nonewline $fd [binary format i $nbins]

    set profileinfoelt  [$profile selectNodes {profileinfo}]
    set profileclass    [$profile getAttribute class]

    switch -exact $profileclass {
        itimer -
        PAPI {
            set profileeventelt [$profile getElementsByTagName profileevent]

            # This produces the inverse of the threshold in gprof output
            # set threshold [[$profileeventelt selectNodes {threshold}] text]

            # This doesn't work (was trying to be tricky...)
            # set threshold [expr {round($threshold / 10000)}]

            # This is what the gprof info page recommends, but unfortunately
            # you lose any indication of sampling frequency in the output.
            set threshold 1

            set desc      [$profileeventelt getAttribute name]
            set descchar "1"
        }
        profil {
            set threshold "100"
            set desc     "seconds"
            set descchar "s"
        }
        default {
            error "unsupported \"class\" attribute for element\
                   <hwpcprofiledata> : \"$class\""
        }
    }

    # rate of profiling clock
    puts -nonewline $fd [binary format i $threshold]

    # long description
    puts -nonewline $fd [binary format a15 $desc]

    # 1-char desc
    puts -nonewline $fd [binary format a1 $descchar]

    # Dump the histogram
    set nextpc $lowpc
    set sampidx 0
    set lastsampidx [llength $mainsamples]
    set sample [lindex $mainsamples $sampidx]
    set pc "0x[$sample getAttribute pc]"
    set recordcount 0
    for {set bin 0} {$bin < $nbins} {incr bin} {
        if { $DEBUG } {
            puts "PROCESSING BIN $bin, searchpc = $pc, thispc = $nextpc"
        }
        if { [string match $pc $nextpc] } {
          set count [$sample text]
          if { $DEBUG } {
              puts "writing non-zero sample"
          }
          incr sampidx
          if { $sampidx < $lastsampidx } {
              set sample [lindex $mainsamples $sampidx]
              set pc "0x[$sample getAttribute pc]"
          }
        } else {
          set count 0
          if { $DEBUG } {
              puts "ZERO SAMPLE"
          }
        }

        puts -nonewline $fd [binary format s [expr {$count & 0xffff}]]
        incr recordcount
        set nextpc "0x[format {%x} [expr {$nextpc + 2}]]"
    }
    if { $DEBUG } {
        puts "Wrote $recordcount records"
    }
}

proc gmon_write {fd hwpcprofiledata} {
    fconfigure $fd -translation binary
    gmon_preamble $fd

    foreach profile $hwpcprofiledata {
        gmon_writeprofile $fd $profile
    }

    return
}
