# $Id: hwpcreport.tcl,v 1.58 2012/05/11 15:23:27 ruiliu Exp $

############################################################################
#
#                    University of Illinois/NCSA
#                        Open Source License
# 
#     Copyright(C) 2002-2006, 2012, The Board of Trustees of the
#            University of Illinois. All rights reserved.
#
#                            Developed by:             
# 
#                        The PerfSuite Project
#           National Center for Supercomputing Applications
#             University of Illinois at Urbana-Champaign
# 
#                  http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

############################################################################
# Text output reporting for performance counters.
#
# Note that this assumes that the source document conforms to
# the <hwpcreport> DTD.  If not, runtime errors may occur, as
# no validation is done by the parser.
#
# Author: Rick Kufrin
# Date:   12/2002
#
# This file is part of PerfSuite

proc hwpcreport {docRoot source {fd stdout}} {
    global havepapi options

    set profilemode 0
    set hwpceventdata [$docRoot getElementsByTagName hwpceventdata]

    if { $hwpceventdata == "" } {
        set hwpceventdata [$docRoot getElementsByTagName hwpcprofiledata]
        if { $hwpceventdata == "" } {
          error [mc "no hwpceventdata or hwpcprofiledata in XML source"]
        } 
        # @@ These tests are no longer strictly necessary, since
        # the -e option is truly optional now.  They're left in to
        # encourage users to supply things correctly.
        if { ! ( $options(vmonout) || $options(gmonout) ) } {
            if { [info exists options(executable)] } {
                if { ! [file exists $options(executable)] } {
                    error "$options(executable) [mc {does not exist}]"
                }
                if { ! [file readable $options(executable)] } {
                    error "$options(executable) [mc {is not readable}]"
                }
            }
        }
        set profilemode 1
    } else {
        if { $options(vmonout) } {
            error [mc "option --vmon requires XML document with profiling data"]
        }
    } 

    if { [catch {$hwpceventdata getAttribute "class"} class] } {
        error [mc "no performance counter \"class\" attribute in XML source"]
    }

    if { ($class == "PAPI") && ( ! $havepapi ) } {
        error "\n[mc {PAPI support not found (required for this XML doc)}]"
    }

    if { ! ( ($class == "PAPI")   || \
             ($class == "perfmon") || \
             ($class == "profil") || \
             ($class == "null")   || \
             ($class == "itimer") ) } {
        error "[mc {this class of performance measurement is not supported}]\
               ($class)"
    }

    if { ! ( $options(xmlout) || $options(vmonout) || $options(gmonout) || \
             $options(cube) ) } {
        hwpcreportOutputPreamble $source $profilemode $fd

        hwpcreportOutputGeneral $docRoot $fd

        if { [info exists options(resourceReport)] } {
            set resourceXMLData [readxml $options(resourceReport)]
            if { [catch {dom parse $resourceXMLData} resourceRoot]} {
                error $resourceRoot
            }
            $resourceRoot documentElement resourceDoc
            # The 0 is a flag that says don't output overall header.
            resourcereport $resourceDoc $options(resourceReport) $fd 0
        }

        if { ! $options(brief) } {
            hwpcreportOutputSystemInfo $docRoot $fd
        }
    }

    if { $profilemode } {
      source [file join [file dir [info script]] profile.tcl]
      psprocess-profile $docRoot $fd
    } else {
        switch $class {
            null {
                source [file join [file dir [info script]] $class-metrics.tcl]
            }
            PAPI -
            perfmon {
                source [file join [file dir [info script]] counter-metrics.tcl]
            }
            default {
                error "[mc {unsupported performance counter library}] \"$class\""
            }
        }

        psprocess-$class $docRoot $fd

        if { $options(calibrate) } {
            hwpcreportOutputCalibrationInfo $docRoot $fd
        }
    }
}

proc hwpcreportOutputGeneric {docRoot ns {fd stdout}} {
    global env options

    set hwpceventData [$docRoot getElementsByTagName hwpceventdata]
    set eventnodes [$hwpceventData selectNodes {hwpcevent}]

    hwpcreportOutputEventHeader $fd

    set n 1
    foreach event $eventnodes {
        # $name is used later
        set name [$event getAttribute name]
        lappend names $name
        if { [catch {::${ns}::getDescription $name} desc] } {
            set desc "$name ([mc {description not available}])"
        } 
        hwpcreportOutputEvent $n $desc [$event text] $fd
        incr n
    }

    if { ! [info exists names] } {
        # There weren't any events in the report.
        # @@ Should the DTD require at least one?
        puts $fd "No hardware event counts recorded."
        return
    }

    hwpcreportOutputEventIndex $names $fd

    # Create a list of event names and values
    set eventList {}
    foreach event $eventnodes {
        lappend eventList [$event getAttribute name] [$event text]
    }

    set clockspeed [[$docRoot getElementsByTagName clockspeed] text]
    # This can overflow integers, so format to a float.
    set wallclock [$docRoot getElementsByTagName wallclock]
    # We have to guard against a missing wallclock element,
    # which can happen in the _to_PS conversion routines
    if { [llength $wallclock] > 0 } {
        scan [$wallclock text] "%f" wallticks
    } else {
        set wallticks -1
    }
    set wallsecs [expr {$wallticks/($clockspeed * 1.0e6)}]
   
    # @@ Assumes only one element named "levels"
    set ncaches [[$docRoot getElementsByTagName levels] text]
    for {set i 1} {$i <= $ncaches} {incr i} {
        set cachearray($i) 0
    }
    set caches [$docRoot getElementsByTagName cache]
    set cachelist {}
    foreach cache $caches {
        set cachetype [$cache getAttribute type]
        switch -exact $cachetype {
            data -
            unified {
                set level [$cache getAttribute level]
                set linesize [[$cache getElementsByTagName linesize] text]
                set cachearray($level) $linesize
            }
            default {
            }
        }
    }
    
    for {set i 1} {$i <= $ncaches} {incr i} {
        lappend cachelist $cachearray($i)
    }

    # @@ BUG (Tcl 8.4.2?  if no flush here, core dump with -brief
    # option).  This sometimes occurs, sometimes not depending
    # on mods to psprocess routines.  Current version dumps with
    # the flush turned on.
    if { [info exists env(PSPROCESS_FLUSH)] } {
        flush $fd
    }

    set metrics [::${ns}::derivedMetrics generate \
            -events $eventList -clockspeed $clockspeed \
            -wallclock $wallsecs -cacheline $cachelist]

    if { $options(peak) > 0 } {
        foreach metric $metrics {
            set desc [lindex $metric 0]
            if { [string match  -nocase "*mfl?ps*cycles*" $desc] } {
                set peakdesc "% peak FLOPS"
                lappend papimetrics [list $peakdesc \
                    [expr { ( [lindex $metric 1] / double($options(peak)) ) \
                             * 100.0 }]]
                break
            }
        }
    }

    hwpcreportOutputMetrics [$hwpceventData getAttribute domain] \
            [$hwpceventData getAttribute multiplexed] \
            [$hwpceventData getAttribute multiplex_usec ""] \
            $metrics $fd
}

proc lr_write {l r fd} {
    puts $fd [format "%-28s : %-s" $l $r]
}

proc hwpcreportOutputPreamble {source profilemode fd} {
    global psprocessVersion options

    puts $fd "[mc {PerfSuite Hardware Performance Summary Report}]\n"
    lr_write "[mc Version]" "1.0" $fd
    lr_write "[mc Created]" "[clock format [clock seconds]]" $fd
    lr_write "[mc Generator]" "psprocess $psprocessVersion" $fd
    if { [info exists options(resourceReport)] } {
        set source "$source $options(resourceReport)"
    }
    lr_write "[mc {XML Source}]" "$source" $fd
    # @@ -e option for profiling mode not required now.
    if { $profilemode && [info exists options(executable)] } {
        lr_write "[mc Executable]" "$options(executable)" $fd
    }
    puts $fd ""

    return
}

proc hwpcreportOutputGeneral {docRoot fd} {
    puts $fd [mc "Execution Information"]
    
    hwpcreportOutputSeparator $fd
    
    set collector [$docRoot getAttribute generator]
    if { $collector != "" } {
        lr_write "[mc Collector]" $collector $fd
    }

    set annotation [$docRoot getElementsByTagName annotation]
    if { [llength $annotation] > 0 } {
        lr_write "[mc Annotation]" "[string trim [$annotation text]]" $fd
    }
    lr_write "[mc Date]" \
                   "[string trim [[$docRoot getElementsByTagName date] text]]"\
                   $fd
    lr_write "[mc Host]" \
                   "[string trim [[$docRoot getElementsByTagName hostname] text]]" \
                   $fd
    set pid [$docRoot getElementsByTagName pid]
    if { [llength $pid] > 0 } {
        lr_write "[mc {Process ID}]" "[string trim [$pid text]]" $fd
        set thread [$docRoot getElementsByTagName thread]
        lr_write "[mc Thread]" "[string trim [$thread text]]" $fd
    }
    lr_write "[mc User]" \
                   "[string trim [[$docRoot getElementsByTagName username] text]]" \
                   $fd
    set command [$docRoot getElementsByTagName command]
    if { [llength $command] > 0 } {
        lr_write "[mc Command]" "[string trim [$command text]]" $fd
        set md5 [$command getAttribute md5sum ""]
        if { $md5 != "" } {
            lr_write "[mc {MD5 Checksum}]" $md5 $fd
        }
    }
    
    puts $fd ""
}

proc hwpcreportOutputSystemInfo {docRoot fd} {
    puts $fd [mc "Processor and System Information"]
    
    hwpcreportOutputSeparator $fd

    set sysRoot [$docRoot getElementsByTagName machineinfo]

    set cpuRoot [$sysRoot getElementsByTagName cpuinfo]

    lr_write "[mc {Node CPUs}]" \
                   "[string trim [[$sysRoot getElementsByTagName cpucount] text]]" \
                   $fd
    lr_write "[mc Vendor]" \
                   "[string trim [[$cpuRoot getElementsByTagName vendor] text]]" \
                   $fd
    set brandelement [$cpuRoot getElementsByTagName brand]
    if { $brandelement != "" } {
        set brand [string trim [$brandelement text]]
        lr_write "[mc Brand]" "$brand" $fd
    }
    lr_write "[mc CPUID Info]" \
                   "[string trim [[$cpuRoot getElementsByTagName cpuidinfo] text]]" \
                   $fd
    lr_write "[mc {CPU Revision}]" \
                   "[string trim [[$cpuRoot getElementsByTagName revision] text]]" \
                   $fd
    lr_write "[mc {Clock (MHz)}]" \
                   "[string trim [[$cpuRoot getElementsByTagName clockspeed] text]]" \
                   $fd
    lr_write "[mc {Memory (MB)}]" \
                   "[string trim [[$sysRoot getElementsByTagName memoryinfo] text]]" \
                   $fd
    lr_write "[mc {Pagesize (KB)}]" \
                   "[string trim [[$sysRoot getElementsByTagName syspagesize] text]]" \
                   $fd

    puts $fd ""

    set cacheRoot [$cpuRoot getElementsByTagName cacheinfo]
    hwpcreportOutputCacheInfo $cacheRoot $fd

    if { 0 } {
    set tlbRoot [$cpuRoot getElementsByTagName tlbinfo]
    hwpcreportOutputTLBInfo $tlbRoot $fd
    }
}

proc hwpcreportOutputCacheInfo {cacheRoot fd} {
    puts $fd [mc "Cache Information"]
    
    hwpcreportOutputSeparator $fd
    set nlevels [string trim [[$cacheRoot getElementsByTagName levels] text]]
    lr_write "[mc {Cache levels}]" "$nlevels" $fd

    for {set level 1} {$level <= $nlevels} {incr level} {
        hwpcreportOutputSeparator $fd 32 "-"
        puts $fd "[mc Level] $level"
        foreach cache [$cacheRoot selectNodes "cache\[@level=\"$level\"\]"] {
            lr_write "[mc Type]" \
                           "[mc [$cache getAttribute type]]" $fd
            set cachesizeelement [$cache getElementsByTagName size]
            lr_write "[mc Size] ([$cachesizeelement getAttribute units])" \
                           "[string trim [$cachesizeelement text]]" $fd
            set linesizeelt [$cache getElementsByTagName linesize]
            if { [llength $linesizeelt] != 0 } {
                lr_write "[mc {Linesize (B)}]" \
                               "[string trim [$linesizeelt text]]" $fd
            }
            lr_write "[mc Assoc]" \
                           "[string trim [[$cache getElementsByTagName associativity] text]]" \
                           $fd
        }
    }

    puts $fd ""
}

proc hwpcreportOutputTLBInfo {tlbRoot fd} {
    puts $fd "[mc {Translation Lookaside Buffer (TLB) Information}]"
    
    hwpcreportOutputSeparator $fd

    foreach tlb [$tlbRoot getElementsByTagName tlb] {
        lr_write "[mc Type]" "[mc [$tlb getAttribute type]]" $fd
        lr_write "[mc Entries]" \
                       "[mc [string trim [[$tlb getElementsByTagName entries] text]]]" \
                       $fd
        lr_write "[mc Assoc]" \
                       "[mc [string trim [[$tlb getElementsByTagName associativity] getAttribute type]]]" $fd
        hwpcreportOutputSeparator $fd 32 "-"
    }

    puts $fd ""
}

proc hwpcreportOutputCalibration {name value fd} {
    set lineLength 92
    set valueWidth 20

    puts -nonewline $fd $name
    set len [string length $name]

    for {set i [incr len]} {$i < [expr ($lineLength - $valueWidth)]} {incr i} {
        puts -nonewline $fd "."
    }
    puts $fd [format " %${valueWidth}s" $value]
}

proc hwpcreportOutputCalibrationInfo {docRoot fd} {
    puts $fd "[mc {Calibration Estimates (based on 1ms sample frequency\
                   and wall clock time)}]"
    hwpcreportOutputSeparator $fd

    set clockspeedelt [$docRoot getElementsByTagName clockspeed]
    if { $clockspeedelt != "" } {
        set clockspeed0     [lindex $clockspeedelt 0]
        set clockspeed      [$clockspeed0 text]
        set clockspeedunits [$clockspeed0 getAttribute units]
        #puts $fd "Clock speed is $clockspeed $clockspeedunits"
    }
    set wallclockelt [$docRoot getElementsByTagName wallclock]

    if { $wallclockelt != "" } {
        set wallclock0     [lindex $wallclockelt 0]
        scan [$wallclock0 text] "%f" wallclock
        set wallclockunits [$wallclock0 getAttribute units]
        #puts $fd "Wall clock is $wallclock $wallclockunits"
    }

    if { [string match -nocase "mhz" $clockspeedunits] && \
         [string match -nocase "ticks" $wallclockunits] } {
        set totalms [expr {$wallclock/(double($clockspeed) * 1.0e3)}]
        #puts $fd "Total milliseconds are $totalms"
    } else {
        #puts $fd "Unsupported units found in XML document\
                  (in <clockspeed> or <wallclock> element)"
    }

    set eventlist [$docRoot getElementsByTagName hwpcevent]
    foreach event $eventlist {
        scan [$event text] "%f" value
        #puts $fd "In calibration loop: value = $value, totalms = $totalms"
        hwpcreportOutputCalibration \
                 [$event getAttribute name] \
                 [expr {int(ceil($value/$totalms))}] \
                 $fd
    }

    puts $fd ""
}

proc hwpcreportOutputSeparator {fd {len 92} {sep "="}} {
    for {set i 0} {$i < $len} {incr i} {
        puts -nonewline $fd $sep
    }
    puts $fd ""
}

proc hwpcreportOutputEventHeader {fd} {
    puts $fd [format "[mc Index] [mc Description] %74s" [mc "Counter Value"]]
    hwpcreportOutputSeparator $fd
}

proc hwpcreportOutputEvent {number desc value fd} {
    set lineLength 86
    set eventWidth 20

    puts -nonewline $fd [format "%5d %s" $number $desc]
    set len [string length $desc]

    for {set i [incr len]} {$i < [expr ($lineLength - $eventWidth)]} {incr i} {
        puts -nonewline $fd "."
    }
    puts $fd [format " %${eventWidth}s" $value]
}

proc hwpcreportOutputEventIndex {names fd} {
    puts $fd ""

    puts $fd [mc "Event Index"]
    hwpcreportOutputSeparator $fd

    set i 1
    foreach name $names {
        puts -nonewline $fd [format "%5d: %-16s" $i $name]
        if { [expr {($i % 4)}] == 0 } {
            puts $fd ""
        }
        incr i
    }

    if { [expr {[llength $names] % 4}] != 0 } {
        puts $fd ""
    }

    puts $fd ""
}

proc hwpcreportOutputMetricDescription {desc fd} {
    set len [string length $desc]
    incr len
    puts -nonewline $fd $desc
    for {set i $len} {$i < 72} {incr i} {
        puts -nonewline $fd "."
    }
}

proc hwpcreportOutputMetric {desc fd} {
    puts $fd [format " %20.3f" $desc]
}

proc hwpcreportOutputMetrics {domain mpx mpx_usec list fd} {
    puts $fd [mc "Statistics"]
    hwpcreportOutputSeparator $fd
    hwpcreportOutputMetricDescription [mc "Counting domain"] $fd
    puts $fd [format " %20s" $domain]
    hwpcreportOutputMetricDescription [mc "Multiplexed"] $fd
    puts $fd [format " %20s" $mpx]
    if { $mpx_usec != "" } {
        hwpcreportOutputMetricDescription [mc "Multiplex interval (microseconds)"] $fd
        puts $fd [format " %20s" $mpx_usec]
    }
    foreach metric $list {
        hwpcreportOutputMetricDescription [lindex $metric 0] $fd
        hwpcreportOutputMetric            [lindex $metric 1] $fd
    }
    puts $fd ""
}
