# $Id: multihwpcreport.tcl,v 1.44 2009/10/02 19:21:18 rkufrin Exp $

############################################################################
#
#                    University of Illinois/NCSA
#                        Open Source License
# 
#          Copyright(C) 2003-2006, The Board of Trustees of the
#              University of Illinois. All rights reserved.
#
#                           Developed by:             
# 
#                      The PerfSuite Project
#             National Center for Supercomputing Applications
#               University of Illinois at Urbana-Champaign
# 
#                  http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

############################################################################
#
# Generate summary performance reports from a collection of
# raw hwpcreports contained in a <multihwpcreport> XML document.
#
# Basic skeleton:
#
# Preamble remains the same as with a single report, but
# there are multiple raw XML source files to list.
#
# Execution Information:
# Earliest and latest dates of completion (latest date done)
# Multiple hostnames (only include unique - done) 
#
# Processor and system information
# Total node CPU counts
# Unique vendor lists
# Unique family lists
# Fastest and slowest clock rates
# Total memory
# Unique pagesizes
#
# No Cache Information section (or aggregate)
#
# Event listings:
# Index Description    Minimum (host) Maximum (host) Mean Median Stddev
#
# Event Index stays the same  (should there be a requirement that
# events are all the same?)
#
# Statistics section split between min/max/mean/median/stddev (done)
#
# Aggregate statistics added: MFLOPS (cyc/wall) MIPS (cyc/wall)
# CPU time total - maybe % peak
# 
# Author: Rick Kufrin
# Date:   2/2003
#
# This file is part of PerfSuite

proc findalleventsinmulti { docRoot } {
    set eventlist {}
    set allevents [$docRoot getElementsByTagName hwpcevent]

    if { $allevents == "" } {
        return {}
    }

    foreach event $allevents {
        set eventname [$event getAttribute name]
        if { [lsearch -exact $eventlist $eventname] == -1 } {
            lappend eventlist $eventname
        }
    }

    return [lsort $eventlist]
}


proc multihwpcreport {docRoot source {fd stdout}} {
    global options havepapi

    set hwpcprofiledata [$docRoot getElementsByTagName hwpcprofiledata]
    if { $hwpcprofiledata != "" } {
        if { $options(vmonout) || $options(cube) } {
            source [file join [file dirname [info script]] profile.tcl]
            if { $options(vmonout) } {
                # Bring in vmon support
                source [file join [file dirname [info script]] vmon.tcl]
                foreach profile $hwpcprofiledata {
                    remove_sharedlibs $profile
                }
                vmon_write $fd $hwpcprofiledata
            } elseif { $options(cube) } {
                source [file join [file dirname [info script]] cube.tcl]
                set profs [$docRoot selectNodes hwpcreport]
                #puts "CUBE: found [llength $profs] hwpcreport nodes"
                if { ! [info exists options(executable)] } {
                    set options(executable) ""
                }
                set cube [::cube::new]
                foreach prof $profs {
                    set profiledata [$prof selectNodes hwpcprofiledata]
                    set proftree [createmappedprofile $profiledata \
                                  $options(executable) $options(showprogress)]
                    #puts $proftree
                    #puts "*******************************"
                    cube_append $cube $proftree $prof
                }
                if { [info exists options(outputFile)] } {
                    set fd $options(outputFile)
                }

                ::cube::dump $cube $fd
                ::cube::delete $cube
            }
            return
        } else {
            # @@ Here is where we could generate derived metrics
            # and output on a module, file, function, line basis.
            # However, another alternative is to work with preprocessed
            # reports individually and let the user decide which
            # regions are of interest. This is what the hwpcfilter.tcl
            # example program contained in the distribution does.
            error "document contains profiling data\
                   (only vmon or Cube3 output are currently supported)"
        }
    }

    if { $options(xmlout) } {
        error "XML output for <multihwpcreport> documents not implemented"
    }

    set eventlist [findalleventsinmulti $docRoot]
    if { [llength $eventlist] == 0 } {
        error "no hwpcevent elements found in $source"
    }

    if { $options(csv) } {
        set delimiter ","
        generatecsv $docRoot $eventlist $fd $delimiter
        return
    }

    set reportlist [$docRoot getElementsByTagName hwpcreport]

    foreach report $reportlist {
        set hwpceventData [$report getElementsByTagName hwpceventdata]
        set class [string tolower [$hwpceventData getAttribute class]]
        # @@ Not necessary? Still need class value
        if { ( $class == "papi" ) && ( ! $havepapi ) } {
            error "PAPI support library not found"
        }
        set eventnodes [$hwpceventData selectNodes {hwpcevent}]

        set eventList {}
        foreach event $eventnodes {
            lappend eventList [$event getAttribute name] [$event text]
        }

        set hostnode [$report getElementsByTagName hostname]
        if { $hostnode != "" } {
            set hostname [$hostnode text]
        }

        set clockspeed [[$report getElementsByTagName clockspeed] text]
        scan [[$report getElementsByTagName wallclock] text] "%f" wallticks
        set wallsecs [expr {$wallticks/($clockspeed * 1.0e6)}]

        set ncaches [[$report getElementsByTagName levels] text]
        for {set i 1} {$i <= $ncaches} {incr i} {
            set cachearray($i) 0
        }
        set caches [$report getElementsByTagName cache]
        set cachelist {}
        foreach cache $caches {
            set cachetype [$cache getAttribute type]
            switch -exact $cachetype {
                data -
                unified {
                    set level [$cache getAttribute level]
                    set linesize [[$cache getElementsByTagName linesize] text]
                    set cachearray($level) $linesize
                }
                default {
                }
            }
        }
        
        for {set i 1} {$i <= $ncaches} {incr i} {
            lappend cachelist $cachearray($i)
        }

        # @@ Note that hostname can appear multiple times.  Fix.
        catch {array unset hostmetrics}
        array set hostmetrics [eval concat [::${class}::derivedMetrics generate \
                -events $eventList -clockspeed $clockspeed \
                -wallclock $wallsecs -cacheline $cachelist]]
        foreach metric [array names hostmetrics] {
            lappend allmetrics($metric) [list $hostname $hostmetrics($metric)]
            # @@ Should be moved to a "real" derived metrics
            # portion for each tcllib/pshwpc subpackage
            if { ( $options(peak) != 0 ) && 
                 ( [string match -nocase "*mfl?ps*cycles*" $metric] ) } {
                set peakindex "% peak FLOPS"
                lappend allmetrics($peakindex) [list $hostname \
                    [expr { ($hostmetrics($metric) / double($options(peak)) ) \
                             * 100.0 }]] 
            }
        }
    }

    hwpcreportOutputPreamble $source 0 $fd

    multihwpcreportOutputGeneral $docRoot $fd

    set maxval 0.0
    foreach metric [array names allmetrics] {
        set sorted($metric) [lsort -real -index 1 $allmetrics($metric)]
        set sum($metric) 0.0
        foreach val $sorted($metric) {
            set sum($metric) [expr {$sum($metric) + [lindex $val 1]}]
        }
        if { $maxval < [expr {abs($sum($metric))}] } {
            set maxval [expr {abs($sum($metric))}]
        }
    }

    set maxhostnamelen [maxhostnamelength $docRoot]
    # Determine how many digits to the left of decimal point needed.
    if { $maxval < 1.0 } {
        set digits 1
    } else {
        set digits [expr {int(floor(log10($maxval)) + 1)}]
    }
    set precision 2
    set width [expr {$digits + $precision + 2}]
    set hostwidth [expr {$maxhostnamelen + 3}]
    set labelwidth [expr {$width + $hostwidth}]
    set fmt [format "%%%d.%df \[%%%ds\]" $width $precision $maxhostnamelen]

    # 92 is the current default complete line length (set in hwpcreport.tcl)
    set textspace [expr {92 - ($width * 2) - ($hostwidth * 2)}]

    puts $fd [format "%-${textspace}s%${labelwidth}s%${labelwidth}s" \
                     [mc "Minimum and Maximum"] [mc "Min"] [mc "Max"]]

    hwpcreportOutputSeparator $fd

    foreach metric [lsort [array names allmetrics]] {
        set nsamples [llength $sorted($metric)]
        puts -nonewline $fd [mc "$metric"]
        if { [string length $metric] > $textspace } {
            puts -nonewline $fd [format "\n%${textspace}s" ""]
        } else {
            for {set i [string length $metric]} {$i < $textspace} {incr i} {
                puts -nonewline $fd "."
            }
        }

        # Include hostname in output for these
        # Min
        puts -nonewline $fd \
                "[format $fmt [lindex [lindex $sorted($metric) 0] 1]\
                              [lindex [lindex $sorted($metric) 0] 0]]"

        # Max
        puts $fd \
                "[format $fmt [lindex [lindex $sorted($metric) end] 1]\
                              [lindex [lindex $sorted($metric) end] 0]]"
    }
    puts $fd ""

    # 92 is the current default complete line length (set in hwpcreport.tcl)
    set fmt [format "%%%d.%df" $width $precision]
    set textspace [expr {92 - ($width * 4)}]

    puts $fd [format "%-${textspace}s%${width}s%${width}s%${width}s%${width}s" \
            [mc "Aggregate Statistics"] [mc "Median"] [mc "Mean"] [mc "StdDev"] [mc "Sum"]]

    hwpcreportOutputSeparator $fd
    
    foreach metric [lsort [array names allmetrics]] {
        set nsamples [llength $sorted($metric)]
        puts -nonewline $fd [mc "$metric"]
        if { [string length $metric] > $textspace } {
            puts -nonewline $fd [format "\n%${textspace}s" ""]
        } else {
            for {set i [string length $metric]} {$i < $textspace} {incr i} {
                puts -nonewline $fd "."
            }
        }

        # Median 
        set mid [expr {$nsamples / 2}]
        if { [expr {$nsamples % 2}] == 0 } {
            set median [expr {([lindex [lindex $sorted($metric) $mid] 1] + \
                               [lindex [lindex $sorted($metric) \
                                         [expr {$mid - 1}]] 1]) / 2.0}]
            puts -nonewline $fd "[format $fmt $median]"
        } else {
            puts -nonewline $fd \
                    "[format $fmt [lindex [lindex $sorted($metric) $mid] 1]]"
        }

        # Mean
        set mean [expr {$sum($metric) / $nsamples}]
        puts -nonewline $fd "[format $fmt $mean]"

        # Standard Deviation
        if { [llength $sorted($metric)] < 2 } {
            set sigma 0.0
        } else {
            set sumsquares 0.0
            foreach val $sorted($metric) {
                set deviation [expr {[lindex $val 1] - $mean}]
                set deviationsquared [expr {$deviation * $deviation}]
                set sumsquares [expr {$sumsquares + $deviationsquared}]
            }
            set sigma [expr {sqrt($sumsquares/($nsamples - 1)) }]
        }
        puts -nonewline $fd "[format $fmt $sigma]"

        # Sum
        puts $fd "[format $fmt $sum($metric)]"
    }
}

proc multihwpcreportOutputGeneral {docRoot fd} {
    puts $fd [mc {Execution Information}]
    
    hwpcreportOutputSeparator $fd
    
    set annotation [$docRoot getElementsByTagName annotation]
    if { [llength $annotation] > 0 } {
        lr_write "[mc Annotation]" \
                    [string trim [[lindex $annotation 0] text]] $fd
    }

    set alldates [$docRoot getElementsByTagName date]
    if { $alldates != "" } {
        set latestdate 0
        foreach dateelt $alldates {
            set thisdate [clock scan [$dateelt text]]
            if { $thisdate > $latestdate } {
                set latestdate $thisdate
            }
        }
    }
    lr_write [mc Date] "[clock format $latestdate]" $fd

    dumpuniquelist $docRoot [mc "Hosts"] hostname $fd
    dumpuniquelist $docRoot [mc "Users"] username $fd

    puts $fd ""
}

proc generateuniquelist {docRoot tag} {
    set all [$docRoot getElementsByTagName $tag]
    set alllist {}
    if { $all != "" } {
        foreach elt $all {
            set thiselt [$elt text]
            if { [lsearch $alllist $thiselt] == -1 } {
                lappend alllist $thiselt
            }
        }
    }
    return $alllist
}

proc dumpuniquelist {docRoot title tag fd} {
    set all [$docRoot getElementsByTagName $tag]
    if { $all != "" } {
        set alllist {}
        foreach elt $all {
            set thiselt [$elt text]
            if { [lsearch $alllist $thiselt] == -1 } {
                lappend alllist $thiselt
            }
        }
        set lastpos 90
        set rootpos 28
        set pos $rootpos
        puts -nonewline $fd [format "%-[expr {$pos + 1}]s:" [mc $title]]
        foreach entry [lsort $alllist] {
            set len [expr {[string length $entry] + 1}]
            if { [expr {$len + $pos}] > $lastpos } {
                puts -nonewline $fd [format "\n%-[expr {$rootpos + 2}]s" ""]
                set pos $rootpos
            }
            puts -nonewline $fd " $entry"
            incr pos $len
        }
        puts $fd ""
    }
}

proc maxhostnamelength {docRoot} {
    set all [generateuniquelist $docRoot hostname]
    set maxlen -1
    foreach host $all {
        set len [string length $host]
        if { $len > $maxlen } {
            set maxlen $len
        }
    }
    return $maxlen
}

# @@ This should be shortened considerably.  Much of the
# code is duplicated from a standalone processor used for
# test purposes and is similar to what is calculated in
# the standard report above.

proc generatecsv { root eventlist fd delimiter {omitheader 0} } {
    global xmlFile options havepapi
    set initialized 0

    set reportlist [$root getElementsByTagName hwpcreport]
    #puts "Found [llength $reportlist] hwpcreports"

    foreach report $reportlist {
        set hwpceventData [$report getElementsByTagName hwpceventdata]
        # @@ This check is duplicated from non-csv - should be
        # placed in a separate proc.
        set class [string tolower [$hwpceventData getAttribute class]]
        # @@ Not necessary?  Still need class value
        if { ( $class == "papi" ) && ( ! $havepapi ) } {
            error "PAPI support library not found"
        }
        if { [info exists options(annotationString)] } {
            set annotation "$options(annotationString)"
        } else {
            set annotation [$report getElementsByTagName annotation]
            if { [llength $annotation] > 0 } {
                set annotation [string trim [[lindex $annotation 0] text]]
            } else {
                set annotation $xmlFile
            }
        }
        set eventnodes [$hwpceventData selectNodes {hwpcevent}]

        set eventList {}
        foreach event $eventnodes {
            lappend eventList [$event getAttribute name] [$event text]
        }

        set clockspeed [[$report getElementsByTagName clockspeed] text]
        scan [[$report getElementsByTagName wallclock] text] "%f" wallticks
        set wallsecs [expr {$wallticks/($clockspeed * 1.0e6)}]

        set ncaches [[$report getElementsByTagName levels] text]
        for {set i 1} {$i <= $ncaches} {incr i} {
            set cachearray($i) 0
        }
        set caches [$report getElementsByTagName cache]
        set cachelist {}
        foreach cache $caches {
            set cachetype [$cache getAttribute type]
            switch -exact $cachetype {
                data -
                unified {
                    set level [$cache getAttribute level]
                    set linesize [[$cache getElementsByTagName linesize] text]
                    set cachearray($level) $linesize
                }
                default {
                }
            }
        }
    
        for {set i 1} {$i <= $ncaches} {incr i} {
            lappend cachelist $cachearray($i)
        }

        set metrics [::${class}::derivedMetrics generate \
                -events $eventList -clockspeed $clockspeed \
                -wallclock $wallsecs -cacheline $cachelist]
        if { ! $initialized } {
            if { ! $omitheader } {
                puts -nonewline $fd "\"ID\"$delimiter"
                foreach event $eventlist {
                    puts -nonewline $fd "$event$delimiter"
                }
            }
            set nmetrics [llength $metrics]
            set metriclist {}
            set count 0
            foreach metric $metrics {
                incr count
                if { ! $omitheader } {
                    #puts -nonewline $fd "[string map {\  _ % percent ( "" ) ""} [lindex $metric 0]]"
                    puts -nonewline $fd "\"[lindex $metric 0]\""
                    if { $count < $nmetrics } {
                        puts -nonewline $fd "$delimiter"
                    }
                }
                lappend metriclist [lindex $metric 0]
            }
            if { ! $omitheader } {
                puts $fd ""
            }
            set initialized 1
        }

        array unset eventarray
        array unset metricarray
        array set eventarray $eventList
        array set metricarray [eval concat $metrics]

        # @@ Assumes $annotation is a string (not a list owing
        # to multiple annotation elements)
        puts -nonewline $fd \"$annotation\"$delimiter
        foreach event $eventlist {
            puts -nonewline $fd $eventarray($event)$delimiter
        }
        set count 0
        foreach metric $metriclist {
            incr count
            if { [info exists metricarray($metric)] } {
                puts -nonewline $fd $metricarray($metric)
            }
            if { $count < $nmetrics } {
                puts -nonewline $fd "$delimiter"
            }
        }
        puts $fd ""
    }
}
