# $Id: profile.tcl,v 1.83 2012/05/11 15:23:27 ruiliu Exp $

############################################################################
#
#                    University of Illinois/NCSA
#                        Open Source License
# 
#      Copyright(C) 2003-2009, 2012, The Board of Trustees of the
#             University of Illinois. All rights reserved.
#
#                            Developed by:             
# 
#                       The PerfSuite Project
#           National Center for Supercomputing Applications
#             University of Illinois at Urbana-Champaign
# 
#                  http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

############################################################################
# Profile processing
#
# Author: Rick Kufrin
# Date:   4/2003
#
# This file is part of PerfSuite.

global havepsbfd havepscxxfilt havepsmd5

# Set up for mapping addresses to lines.  First try the
# PerfSuite BFD Tcl extension, if not available
# use the subordinate process-based ones in the pure Tcl
# PerfSuite binutils package.

if { [catch {package require psbfd} res] } {
    if { [catch {package require binutils} res ] } {
        puts stderr "could not find either BFD or binutils support"
        exit 1
    } else {
        set havepsbfd 0
    }
} else {
    set havepsbfd 1
}

# Set up for function name demangling.  First try the
# PerfSuite libiberty C extension, if not available
# fall back to the binutils package and load it if
# not already loaded.

if { [catch {package require pscxxfilt} res] } {
    set havepscxxfilt 0
    if { $havepsbfd } {
        package require binutils
    }
} else {
    set havepscxxfilt 1
} 

# Set up for MD5 checksum generation

if { [catch {package require psmd5} res] } {
    set havepsmd5 0
} else {
    set havepsmd5 1
}

proc psprocess-profile {docRoot {fd stdout}} {
    global options addrmap havepsbfd havepscxxfilt havepsmd5 env

    set hwpcprofileData [$docRoot getElementsByTagName hwpcprofiledata]
    set totalsampleselt [$hwpcprofileData selectNodes {totalsamples}]
    if { $totalsampleselt == "" } {
        error "no XML element \"totalsamples\" found"
    } else {
        set totalSamples [$totalsampleselt text]
    }

    if { $options(vmonout) || $options(gmonout) } {
        source [file join [file dirname [info script]] vmon.tcl]
        if { $options(gmonout) } {
            source [file join [file dirname [info script]] gmon.tcl]
        }
        remove_sharedlibs $hwpcprofileData
        if { $options(vmonout) } {
            vmon_write $fd $hwpcprofileData
        }
        if { $options(gmonout) } {
            puts stderr "Note: gprof may require the option --flat-profile\
                         to process $options(outputFile)"
            gmon_write $fd $hwpcprofileData
        }
        return
    }

    # Ensure that options(executable) can be referenced.
    # @@ Should address by referencing options as a global array.
    if { ! [info exists options(executable)] } {
        set options(executable) ""
    }

    # If the "-e" option was supplied and an MD5 checksum is available
    # in the XML document, check that the checksum matches with
    # the executable given.  If not, warn to standard error.
    if { $havepsmd5 && ( $options(executable) != "" ) } {
        if { ! [catch {::psmd5::md5 -file $options(executable)} exesum] } {
            set cmdelts [$docRoot getElementsByTagName command]
            if { $cmdelts != "" } {
                set cmdelt [lindex $cmdelts 0]
                set cmdsum [$cmdelt getAttribute md5sum ""]
                if { $cmdsum != "" } {
                    if { ! [string equal $cmdsum $exesum] } {
                        # Save this for later, so that the user sees
                        # it after the report when not redirecting to a file.
                        set md5warn [list $cmdsum $exesum]
                    }
                }
            }
        }
    }

    if { $options(cube) } {
        source [file join [file dirname [info script]] cube.tcl]
        set proftree [createmappedprofile $hwpcprofileData $options(executable) \
                          $options(showprogress)]

        set cube [::cube::new]
        cube_append $cube $proftree $docRoot

        if { [info exists options(outputFile)] } {
            set fd $options(outputFile)
        }
        
        ::cube::dump $cube $fd

        ::cube::delete $cube

        return
    }

    if { [catch {mapaddrs $hwpcprofileData $options(executable) addrmap \
                     $options(showprogress)} res] } {
        error $res
    }

    foreach addr [array names addrmap] {
        #puts "$addr: $addrmap($addr)"
        set samples  [lindex $addrmap($addr) 0]
        set module   [lindex $addrmap($addr) 1]
        set function [lindex $addrmap($addr) 2]
        set file     [lindex $addrmap($addr) 3]
        set lineno   [lindex $addrmap($addr) 4]
        set line     $function:$file:$lineno

        if { ! [info exists moduleTotals($module)] } {
            set moduleTotals($module) $samples
        } else {
            incr moduleTotals($module) $samples
        }
        if { ! [info exists functionTotals($function)] } {
            set functionTotals($function) $samples
        } else {
            incr functionTotals($function) $samples
        }
        if { ! [info exists fileTotals($file)] } {
            set fileTotals($file) $samples
        } else {
            incr fileTotals($file) $samples
        }
        if { ! [info exists lineTotals($line)] } {
            set lineTotals($line) $samples
        } else {
            incr lineTotals($line) $samples
        }
    }
    
    # Guard against nothing being set above
    set fileTotalList {}
    set functionTotalList {}
    set functionTotalListDemangled {}
    set lineTotalList {}
    set lineTotalListDemangled {}
    set moduleTotalList {}

    if { $options(xmlout) || $options(showmodules) } {
        foreach module [array names moduleTotals] {
            lappend moduleTotalList [list $module $moduleTotals($module)]
        }
        set moduleTotalList [lsort -decreasing -integer -index 1 \
                                   $moduleTotalList]
    }

    if { $options(xmlout) || $options(showfiles) } {
        foreach file [array names fileTotals] {
            lappend fileTotalList [list $file $fileTotals($file)]
        }
        set fileTotalList [lsort -decreasing -integer -index 1 \
                                   $fileTotalList]
    }

    if { $options(xmlout) || $options(showfunctions) } {
        foreach function [array names functionTotals] {
            lappend functionTotalList [list $function \
                                            $functionTotals($function)]
        }
        set functionTotalList [lsort -decreasing -integer -index 1 \
                                   $functionTotalList]
    }

    if { $options(xmlout) || $options(showlines) } {
        foreach line [array names lineTotals] {
            lappend lineTotalList [list $line $lineTotals($line)]
        }
        set lineTotalList [lsort -decreasing -integer -index 1 \
                                   $lineTotalList]
    }

    # @@ Duplicated from hwpcreport.tcl
    set clockspeed [[$docRoot getElementsByTagName clockspeed] text]
    # This can overflow integers, so format to a float.
    scan [[$docRoot getElementsByTagName wallclock] text] "%f" wallticks
    set wallsecs [expr {$wallticks/($clockspeed * 1000000.0)}]
   
    set profileinfoelt  [$hwpcprofileData selectNodes {profileinfo}]
    set profileclass    [$hwpcprofileData getAttribute class]
    if { [catch {$hwpcprofileData getAttribute version} swversion] } {
        set swversion "unknown"
    }

    switch -exact $profileclass {
        itimer -
        PAPI -
        perfmon {
            set profileeventelt [$profileinfoelt selectNodes {profileevent}]
            set eventname       [$profileeventelt getAttribute name]
            set domain          [$profileeventelt getAttribute domain]
            set threshold       [[$profileeventelt selectNodes {threshold}] \
                                 text]
        } 
        profil {
            set profileeventelt "profil"
            set eventname       "milliseconds"
            set domain          "user"
            # @@ The resolution should have been finer for IA-64:
            #    "1" instead of "10".
            #    This is a bad workaround, would be better
            #    done by querying the machine here
            set threshold "10"
        }
        default {
            error "unsupported \"class\" attribute for element\
                   <hwpcprofiledata> : \"$profileclass\""
        }
    }

    array set headerdata [list class $profileclass \
                               version $swversion \
                               event $eventname \
                               period $threshold \
                               samples $totalSamples \
                               domain $domain \
                               wallsecs $wallsecs \
                         ]

    if { ! $options(xmlout) } {
        puts $fd "[mc {Profile Information}]"
        hwpcreportOutputSeparator $fd
        lr_write "[mc Class]" "$headerdata(class)" $fd
        lr_write "[mc Version]" "$headerdata(version)" $fd
        if { [string match $headerdata(class) "PAPI"] } {
            if { [catch {::papi::getDescription $eventname} desc] } {
                set desc "description not available"
            }
            lr_write "[mc Event]" "$headerdata(event) [mc ($desc)]" $fd
        } elseif { [string match $headerdata(class) "itimer"] } {
            switch -exact -- $headerdata(event) {
                ITIMER_PROF {
                    set desc "Process time in user and system mode"
                }
                ITIMER_REAL {
                    set desc "Real time execution"
                }
                ITIMER_VIRTUAL {
                    set desc "Process time in user mode"
                }
                default {
                }
            }
            lr_write "[mc Event]" "$headerdata(event) [mc ($desc)]" $fd
        } else {
            lr_write "[mc Event]" "$headerdata(event)" $fd
        }
        lr_write "[mc Period]" "$headerdata(period)" $fd
        lr_write "[mc Samples]" "$headerdata(samples)" $fd
        lr_write "[mc Domain]" "[mc $headerdata(domain)]" $fd
        lr_write "[mc {Run Time}]" "[format %.2f $headerdata(wallsecs)]\
                                 ([mc seconds])" $fd


        #puts -nonewline $fd "[mc {Min Self}] %   : "
        if { $options(threshold) > 0 } {
            set thresholdstring $options(threshold)
            #puts $fd "$options(threshold)"
        } else {
            set thresholdstring "([mc all])"
            #puts $fd "([mc all])"
        }
        lr_write "[mc {Min Self}] %" "$thresholdstring" $fd
        puts $fd ""

        if { $options(showmodules) } {
            hwpcreportOutputProfileStats $fd "[mc Module]" $moduleTotalList \
                $totalSamples
        }
        if { $options(showfiles) } {
            hwpcreportOutputProfileStats $fd "[mc File]" $fileTotalList \
                $totalSamples
        }
        if { $options(showfunctions) || $options(showlines) } {
            if { $havepscxxfilt } {
                set demangleproc ::pscxxfilt::demangle
            } else {
                ::binutils::cxxfilt::open
                set demangleproc ::binutils::cxxfilt::filter
            }
        }
        if { $options(showfunctions) } {
            foreach elt $functionTotalList {
                lappend functionTotalListDemangled \
                    [list [$demangleproc [lindex $elt 0]] [lindex $elt 1]]
            }
            hwpcreportOutputProfileStats $fd "[mc Function]" \
                $functionTotalListDemangled $totalSamples
        }
        if { $options(showlines) } {
            foreach elt $lineTotalList {
                set tmp [split [lindex $elt 0] :]
                set dm [$demangleproc [lindex $tmp 0]]
                set eltDM [join [list $dm [lindex $tmp 1] [lindex $tmp 2]] :]
                lappend lineTotalListDemangled [list $eltDM [lindex $elt 1]]
            }
            hwpcreportOutputProfileStats $fd \
                "[mc Function]:[mc File]:[mc Line]" \
                $lineTotalListDemangled $totalSamples
        }
        if { ! $havepscxxfilt && \
             ( $options(showfunctions) || $options(showlines) ) } {
            ::binutils::cxxfilt::close
        }
    } else {
        source [file join [file dirname [info script]] profilexml.tcl]
        set profileXML [ORIGhwpcreportCreateProfileXML \
		            $docRoot headerdata \
                            $moduleTotalList $fileTotalList \
                            $functionTotalList functionTotals \
                            $lineTotalList lineTotals]

        # Emit the XML preamble.  tDOM does not do this for you.
        puts $fd "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>"

        # Only attempt to output the DTD in the result if it was set
        # in the hwpcreportCreateProfileXML procedure.
        if { [$profileXML systemId] != "" } {
            puts $fd "[$profileXML asXML -indent 2 -doctypeDeclaration true]"
        } else {
            puts $fd "[$profileXML asXML -indent 2]"
        }
        $profileXML delete
    }

    if { [info exists md5warn] } {
        puts stderr "Warning: MD5 checksums do not match:"
        puts stderr "\t\"[lindex $md5warn 0]\"\tXML"
        puts stderr "\t\"[lindex $md5warn 1]\"\t$options(executable)"
        unset md5warn
    }
}

# mapmodule: map the samples corresponding to an individual module
#
# This is placed in a separate procedure to allow for byte-compilation

proc mapmodule {modulenode exe m {showProgress 0} {progressFrequency 0} \
                {tracemap 0} {tracefd ""}} {

    global havepsbfd samplecount
    upvar $m map

    set samplenodes [$modulenode selectNodes {samples/sample}]
    if { [llength $samplenodes] == 0 } {
        if { $tracemap } {
            puts $tracefd "# No samples in module\
                          [$modulenode getAttribute file]"
        }
        return
    }

    set filename [$modulenode getAttribute file]
    set offset [$modulenode getAttribute offset]

    # First we try to open the file listed in the
    # XML document.  If not available, try the
    # executable passed in (if any) next.
    # @@ It is possible that we could attempt to
    # use the $exe file for an inaccessible module.

    if { ! [file exists $filename] } {
        puts stderr "Warning: the following module contained in XML document\
                     does not exist:"
        puts stderr "\t\"$filename\""

        if { [file exists $exe] } {
            puts stderr "Replacing with user-supplied executable: \"$exe\""
            set filename $exe
        } else {
            puts stderr "Samples attributed to this module will be ignored."
            return
        }
    }

    if { $havepsbfd } {
        ::psbfd::open $filename
        set sharedlib [::psbfd::inquire dynamic]
    } else {
        ::binutils::addr2line::open $filename
        set sharedlib [::binutils::file::inquire dynamic $filename]
    }
    if { $tracemap } {
        puts $tracefd "# module: $filename"
    }
    
    foreach sampleelt $samplenodes {
        if { $showProgress } {
            incr samplecount
            # @@ Why not just test for equality?
            if { [expr {$samplecount % $progressFrequency}] == 0 } {
                puts -nonewline stderr "."
                set samplecount 0
            }
        }

        # address     - original address in process space
        # fileaddress - adjusted address (same as original
        #               for non-shared, adjusted for shared)
        # Index into map is $address
        
        set address [$sampleelt getAttribute pc]

        if { $tracemap } {
            puts -nonewline $tracefd "$address "
        }

        # @@ The expr here in setting the address is a potential problem
        # source with parsing and expression.
        if { $sharedlib } {
            set tmpaddress "0x$address"
            set tmpoffset  "0x$offset"
            set tmpaddress [expr {round(double($tmpaddress) - \
                                        double($tmpoffset))}]
            set fileaddress [format "%lx" $tmpaddress]
        } else {
            set fileaddress $address
        }

        # @@ If this address already has an entry in the map,
        # an error should be flagged... should not happen.

        if { ! [info exists map($address)] } {
            # @@ Should be able to remove this conditional
            # through appropriate renaming of the proc.
            # Namespace variables will have to be checked for
            # absolute scope though.
            if { $havepsbfd } {
                if { [catch {::psbfd::lookup $fileaddress} res] } {
                    set map($address) [list [$sampleelt text] \
                                             $filename ?? ?? 0]
                } else {
                    set map($address) [concat [$sampleelt text] \
                                               $filename $res]
                }
            } else {
                set map($address) [concat [$sampleelt text] $filename \
                                       [::binutils::addr2line::lookup \
                                            $fileaddress]]
            }

            if { $tracemap } {
                puts $tracefd "$fileaddress $map($address)"
            }
        }
    }

    if { $havepsbfd } {
        ::psbfd::close
    } else {
        ::binutils::addr2line::close
    }
}

#
# mapaddrs: populate the array addrmap with a mapping of program
#           addresses to source code.
#
# On successful return, each array element will be indexed by
# the hex address and contain a five-item list corresponding
# to that address:
#
#           { samples module function file line }
#
# If information about the function and file is not available,
# their slots will contain "??".  Unknown line numbers will
# be reported as "0".

proc mapaddrs {profileData exe addrmap {showProgress 0}} {
    global havepsbfd env timers samplecount
    upvar $addrmap map 
    set samplecount 0
    set progressFrequency 100

    set timers(start,bfd) [clock clicks -milliseconds]

    # Allow the user to force use of addr2line, even if PerfSuite BFD
    # extension is available.  This is in effect only for the lifetime
    # of this procedure; original behavior is restored before return.

    set havepsbfdorig $havepsbfd

    if { [info exists env(PSPROCESS_MAPPER)] } {
        switch -exact $env(PSPROCESS_MAPPER) {
            addr2line {
                set havepsbfd 0
                package require binutils
            }
            default {
                puts stderr "Warning: invalid value for PSPROCESS_MAPPER:\
                            \"$env(PSPROCESS_MAPPER)\" (ignored)"
            }
        }
    }

    # Set up for debugging output based on value of environment variable
    # PSPROCESS_TRACE.  If set, output PC->source line mappings there.

    set tracemap 0
    set tracefd ""
    if { [info exists env(PSPROCESS_TRACE)] } {
        if { [catch {open $env(PSPROCESS_TRACE) "w"} tracefd] } {
            puts stderr "Warning: cannot open profiling trace map: $tracefd"
        } else {
            set tracemap 1
            puts $tracefd "# psprocess trace map opened at\
                           [clock format [clock seconds]]"
            puts $tracefd "# executable: $exe"
            puts -nonewline $tracefd "# mapper: "
            if { $havepsbfd } {
                puts $tracefd "psbfd"
            } else {
                puts $tracefd "addr2line"
            }
            puts $tracefd "# PC file-offset samples module function file line"
            puts $tracefd "#"
        }
    }

    set modulenodes [$profileData selectNodes module]

    if { $showProgress } {
        puts -nonewline stderr "[mc {Processing samples}]"
    }

    foreach modulenode $modulenodes {
        mapmodule $modulenode $exe map $showProgress $progressFrequency \
                  $tracemap $tracefd
    }

    if { $showProgress } {
        puts stderr ""
    }

    # Restore original setting of BFD flag, close debugging output
    # if in use.

    set havepsbfd $havepsbfdorig

    if { $tracemap } {
        close $tracefd
    }

    set timers(stop,bfd) [clock clicks -milliseconds]
    set timers(total,bfd) [expr {$timers(stop,bfd) - $timers(start,bfd)}]
}

proc hwpcreportOutputProfileStats {fd title samplelist totalsamples} {
    global options

    set linelength 80
    set samplewidth [expr {[string length $totalsamples] + 2}]

    puts $fd "$title [mc Summary]"
    hwpcreportOutputSeparator $fd $linelength "-"
    puts -nonewline $fd [format %${samplewidth}s [mc Samples]]
    puts $fd [format "%9s%9s  %-s" "[mc Self] %" "[mc Total] %" $title]
    puts $fd ""

    set cumsamples 0
    foreach elt $samplelist {
        set samples [lindex $elt 1]
        incr cumsamples $samples
        set pct [expr {(double($samples)/$totalsamples) * 100.0}]
        if { $pct < $options(threshold) } {
            break
        }
        puts $fd [format " %${samplewidth}d %7.2f%% %7.2f%%  %-s" \
                      $samples \
                      $pct \
                      [expr {(double($cumsamples)/$totalsamples) * 100.0}] \
                      [lindex $elt 0]]
    }
    puts $fd ""

}

# Delete all profile children that represent
# shared libraries

proc remove_sharedlibs { hwpcprofile } {
    global havepsbfd

    set debug 0
    set modulenodes [$hwpcprofile selectNodes module]

    if { $debug } {
        puts "Entering remove_sharedlibs, module count: [llength $modulenodes]"
        set removed 0
    }

    foreach modulenode $modulenodes {
        set filename [$modulenode getAttribute file]
        # @@ An alternative approach for error-handling may be
        # to completely discard the file in question.  The concern
        # is the potential of having all files discarded, so
        # needs some thought.
        #
        # Because of issues caused by more recent versions of Linux
        # that include modules like "[stack]", which are not even
        # proper files, the error handling has been changed to flag
        # any such instances as "shared libraries", so that they too
        # will be removed from the XML document before processing downstream.
        if { $havepsbfd } {
            if { [catch {::psbfd::open $filename} res] } {
                # error "Could not open \"$filename\": $res"
                set sharedlib 1
                # @@ Not sure what I was thinking here...
                # set filename $exe
                # ::psbfd::open $filename
            } else {
                set sharedlib [::psbfd::inquire dynamic]
                ::psbfd::close
            }
        } else {
            if { [catch {::binutils::addr2line::open $filename} res] } {
                # error "Could not open \"$filename\": $res"
                set sharedlib 1
                # @@ or here...
                # set filename $exe
                # ::binutils::addr2line::open $filename
            } else {
                set sharedlib [::binutils::file::inquire dynamic $filename]
                ::binutils::addr2line::close
            }
        }
        if $sharedlib {
            if { $debug } {
                set count 0
                foreach sample [$modulenode getElementsByTagName sample] {
                    set count [$sample text]
                    incr removed $count
                }
                puts "Removing module $filename ($count samples)"
            }
            $hwpcprofile removeChild $modulenode
        }
    }
    set modulenodes [$hwpcprofile selectNodes module]

    if { $debug } {
        puts "Exiting remove_sharedlibs, module count: [llength $modulenodes]"
        puts "\tFiles:"
        foreach modulenode $modulenodes {
            puts "\t\t[$modulenode getAttribute file]"
        }
        puts "Samples in removed modules: $removed"
    }
}

#
# Main procedure for mapping samples to source code
#
# Input parameters:
#    sampledata =   the <hwpcprofiledata> element in a "hwpcreport" document
#                   generated in profiling mode
#    exe          = name of the executable (intended to be the contents of
#                   psprocess' global options(executable) variable
#    showprogress = 0/1-valued, causes verbose output to be sent to
#
# On success, this procedure returns the entire profile, formatted in the
# "new" profile tree-structured format (see below).

proc createmappedprofile { sampledata exe showprogress } {
    global havepscxxfilt

    if { $havepscxxfilt } {
        set demangleproc ::pscxxfilt::demangle
    } else {
        ::binutils::cxxfilt::open
        set demangleproc ::binutils::cxxfilt::filter
    }

    if { [catch {mapaddrs $sampledata $exe addrmap $showprogress} res] } {
        error $res
    }

    foreach addr [array names addrmap] {
        set samples  [lindex $addrmap($addr) 0]
        set module   [lindex $addrmap($addr) 1]
        set function [$demangleproc [lindex $addrmap($addr) 2]]
        set file     [lindex $addrmap($addr) 3]
        set lineno   [lindex $addrmap($addr) 4]
        set line     $function:$file:$lineno

        profile proftree add \
            -program $exe \
            -module  $module \
            -file    $file \
            -function $function \
            -line     $lineno \
            $samples
    }

    return $proftree
}


############################################################################
#     NEW SUPPORT FOR "TREE"-STRUCTURED PROFILE DATA FORMAT.
############################################################################

# profileadd : recursive procedure to add the given number of samples
#              to the proper portion of a profile (or subprofile) identified
#              by "idlist", which is the module/file/function/line strings.
#
#
# This procedure operates on the data structure passed in as "profile" as
# follows: the list consists of:
#
#  {TAG1 NAME1 SAMPLES1 {SUBPROFILE}} {TAG2 NAME2 SAMPLES2 {SUBPROFILE2}} ? ...?
#                 ^                                 ^
#                 |                                 |
#                nameidx is set to one of these indices
#
#

proc profileadd { profile idlist samples } {
    set debug 0

    # exe is not currently used.
    array set tags {5 exe 4 module 3 file 2 function 1 line}

    # Retrieve the current component to be added/modified (mod/file/func/line)
    set target  [lindex $idlist 0]

    if { $debug } {
        puts "================================================="
        puts "profileadd entry:\n\tidlist: $idlist"
        puts "Profile on entry (length [llength $profile]):\n\t$profile"
    }

    # Examine each of the NAMEx entries of the profile passed in to see if there
    # is already an entry associated with the target name.

    set foundname 0
    set nameidx 0
    foreach elt $profile {
        if { [lindex $elt 1] == $target } {
            set foundname 1
            if { $debug } {
                puts "found \"$target\" at index $nameidx"
                puts "need to work with:\n\t$elt"
            }
            break
        }
        incr nameidx
    }


    # This block takes care of the sample count at this level (associated with
    # the current identifier).

    if { $foundname } {
        set nameelt [lindex $profile $nameidx]
        set oldsamples [lindex $nameelt 2]
        set newsamples [expr {$oldsamples + $samples}]
        set nameelt [lreplace $nameelt 2 2 $newsamples]
        set profile [lreplace $profile $nameidx $nameidx $nameelt]
        if { $debug } {
            puts "Old sample count was: $oldsamples, new: $newsamples"
            puts "New nameelt: [lindex $profile $nameidx]"
        }
    } else {
        set nameelt [list $tags([llength $idlist]) [lindex $idlist 0] $samples]
        if { [llength $idlist] > 1 } {
            lappend nameelt {}
        }
        lappend profile $nameelt
        set nameidx [expr {[llength $profile] -1}]
    }

    # If there are more components to deal with, recurse to handle them.
    # In order to do so, we pass the SUBPROFILE associated with the profile
    # list element that was either found and modified or created above.  That
    # element is pointed to by nameidx.

    if { [llength $idlist] > 1 } {
        set subprofile [lindex [lindex $profile $nameidx] 3]
        set newsub [profileadd $subprofile [lrange $idlist 1 end] $samples]
        set newnameelt [lreplace $nameelt 3 3 $newsub]
        set profile [lreplace $profile $nameidx $nameidx $newnameelt]
        if { $debug } {
            puts "New profile (length [llength $profile]) is:\n\t$profile"
        }
    }

    return $profile
}

#
# Procedure profile: manipulate PerfSuite list-oriented flat profiles.
#
# Synopsis:
#
#    profile proflist subcmd ?args?
#
#    subcmd: get [-program name [-module name [-file name [-function name
#                    [-line number]]]]]
#            add -program name -module name -file name -function name 
#                    -line number val

proc profile { profile cmd args } {
    upvar $profile p
    set debug 0

    if { $debug } {
        puts "profile: $profile cmd: $cmd args: $args"
    }

    switch -- $cmd {
        get -
        add {
            if { $debug } {
                puts "Doing $cmd"
            }
            set argc [llength $args]
            for { set i 0 } { $i < $argc } { incr i } {
                set arg [lindex $args $i]
                switch -- $arg {
                    -program -
                    -module -
                    -file -
                    -function -
                    -line {
                        set opt [string range $arg 1 end]
                        incr i
                        if { $i == $argc } {
                            error "option $arg requires an argument"
                        }
                        set $opt [lindex $args $i]
                    }
                    default {
                        if { $i != [expr $argc - 1] } {
                            error "unrecognized option: $arg"
                        }
                        set val $arg
                    }
                }
            }
        }
        default {
            error "unrecognized subcommand: $cmd"
        }

    }

    if { $debug } {
        puts "Current variables: [info vars]"
        foreach var {program module file function line} {
            puts -nonewline "$var: "
            if { [info exists $var] } {
                set varval [set $var]
                puts $varval
            } else {
                puts "(not set)"
            }
        }
    }

    if { ! ([info exists program] && [info exists module] && \
            [info exists file] && [info exists function] && \
            [info exists line]) } {
        error "must supply values for program, module, function, file,\
               and line"
    }

    if { ! [info exists p] } {
        set p {}
    }

    switch -- $cmd {
        add {
            if { ! [info exists val] } {
                error "no sample count provided"
            }
            set p [profileadd $p \
                             [list $program $module $file $function $line] \
                             $val]
        }
        get {
            error "get subcommand not yet implemented"
        }
    }

    return
}
