# $Id: profilexml.tcl,v 1.9 2009/10/21 18:20:03 rkufrin Exp $

############################################################################
#
#                      University of Illinois/NCSA
#                          Open Source License
# 
#          Copyright(C) 2003-2009, The Board of Trustees of the
#              University of Illinois. All rights reserved.
#
#                             Developed by:             
# 
#                        The PerfSuite Project
#            National Center for Supercomputing Applications
#              University of Illinois at Urbana-Champaign
# 
#                   http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################


# hwpcreportCreateProfileXML
#
# Newer implementation of procedures to build DOM tree for a profile.
#
# This version attempts to modularize the output, using common procs to
# generate DOM nodes that are common to all profile-related output.
#
# In general, the current PerfSuite post-processed profile DTD looks like:
#
# <hwpcprofilereport>
# <date> </date>
# <hostname> </hostname>
# <domainname> </domainname>
# <username> </username>
# <executioninfo>
#   <date> </date>
#   <hostname> </hostname>
#   <domainname> </domainname>
#   <pid> </pid>
#   <thread> </thread>
#   <username> </username>
#   <command> </command>
#   <wallclock> </wallclock>
#   <cputime>                      <- only introduced in 0.2
#     <usertime> </usertime>       <- only introduced in 0.2
#     <systemtime> </systemtime>   <- only introduced in 0.2
#   </cputime>                     <- only introduced in 0.2
#   <machineinfo> </machineinfo>
#     <cpucount> </cpucount>
#     <cpuinfo> </cpuinfo>
#       <vendor> </vendor>
#       <family> </family>
#       <brand> </brand>
#       <revision> </revision>
#       <clockspeed> </clockspeed>
#       <cacheinfo>
#         <levels> </levels>
#         <cache>
#           <size> </size>
#           <associativity> </associativity>
#         </cache>
#         ..... repeated <cache> elements ...
#       </cacheinfo>
#       <tlbinfo>
#         <tlb>
#           <entries> </entries>
#           <associativity> </associativity>
#         </tlb>
#         ..... repeated <tlb> elements ...
#       </tlbinfo>
#     </cpuinfo>
#     <memoryinfo> </memoryinfo>
#     <syspagesize> </syspagesize>
#    </machineinfo>
#  </executioninfo>
#  <hwpcprofile>
#    <hwpcevent> </hwpcevent>
#    <period> </period>
#    <totalsamples> </totalsamples>
#    <module>
#      <file>
#        <function>
#          <line> </line>
#        </function>
#      </file>
#    </module>
#     ... repeated <module> elements ...
#  </hwpcprofile>
#        

# BELOW IS THE ORIGINAL VERSION (still used)
#
# hwpcreportCreateProfileXML: generate an XML DOM document that contains
#                             the results of profiling associated
#                             with the functions/files/lines
#
# @@ This is most certainly an extremely inefficient implementation!
#    The data structures passed in cause too much shuffling and
#    searching - should be redone.
#
# <!---------------------------------------------------->
# <!-- NOTE: this document format is likely to change -->
# <!---------------------------------------------------->
#
# Author: Rick Kufrin
# Date:   5/2003
#
# This file is part of PerfSuite.
#

proc ORIGhwpcreportCreateProfileXML {profileRoot headerData moduleTotalList \
                                     fileTotalList \
                                     functionTotalList functionTotals \
                                     lineTotalList lineTotals} {
    upvar $lineTotals     lines
    upvar $functionTotals functions
    upvar $headerData     header
    global options env tcl_platform havepsbfd havepscxxfilt

    #puts [array names functions]
    #exit 0

    # <hwpcprofilereport>
    set xmldoc [dom createDocument "hwpcprofilereport"]

    # DTD
    if { [info exists env(PS_DTDDIR)] } {
        $xmldoc systemId "$env(PS_DTDDIR)/pshwpc/hwpcprofilereport.dtd"
    }
    $xmldoc documentElement xmlroot
    $xmlroot setAttribute \
        version "0.1" generator "psprocess" \
        xmlns:perfsuite "http://perfsuite.ncsa.uiuc.edu/"

    # <hwpcprofilereport/date>
    $xmlroot appendFromList \
        [list date {} \
             [list [list \#text [clock format [clock seconds] \
                                     -format "%a %b %d %H:%M:%S %Y"]]]]

    # <hwpcprofilereport/hostname>
    set hostname (null)
    foreach index {HOSTNAME hostname} {
        if [info exists env($index)] {
            set hostname $env($index)
            break
        }
    }
    $xmlroot appendFromList \
        [list hostname {} \
             [list [list \#text $hostname]]]

    # <hwpcprofilereport/domainname> (faked)
    $xmlroot appendFromList \
        [list domainname {} \
             [list [list \#text (null)]]]

    # <hwpcprofilereport/username>
    $xmlroot appendFromList \
        [list username {} \
             [list [list \#text $tcl_platform(user)]]]

    # <hwpcprofilereport/annotation>
    if { [info exists options(annotationString)] } {
        $xmldoc createElement annotation annotationObj
        $xmldoc createTextNode $options(annotationString) annotationTextObj
        $annotationObj appendChild $annotationTextObj
        $xmlroot appendChild $annotationObj
    }

    # <hwpcprofilereport/executioninfo>
    $xmldoc createElement executioninfo exeInfoObj
    $xmlroot appendChild $exeInfoObj

    # <hwpcprofilereport/executioninfo/<various>
    foreach element {date annotation hostname domainname pid thread username \
                     command wallclock} {
        set subtree [$profileRoot getElementsByTagName $element] 
        if { $subtree != "" } {
            $exeInfoObj appendFromList [$subtree asList]
        }
    }

    # <hwpcprofilereport/executioninfo/machineinfo
    $exeInfoObj appendFromList \
        [[$profileRoot getElementsByTagName machineinfo] asList]

    # <hwpcprofilereport/hwpcprofile>
    $xmldoc createElement hwpcprofile profileObj
    if { $havepsbfd } {
        $profileObj setAttribute mapper psbfd
    } else {
        $profileObj setAttribute mapper binutils
    }

    # <hwpcprofilereport/hwpcprofile/hwpcevent>
    # @@ preset hardcoded - change
    $profileObj appendFromList \
        [list hwpcevent [list class $header(class) \
                             version $header(version) \
                             type preset name $header(event) \
                             domain $header(domain)] {}]

    # <hwpcprofilereport/hwpcprofile/period>
    $profileObj appendFromList \
        [list period {} \
             [list [list \#text $header(period)]]]

    # <hwpcprofilereport/hwpcprofile/totalsamples>
    $profileObj appendFromList \
        [list totalsamples {} \
             [list [list \#text $header(samples)]]]

    # <hwpcprofilereport/hwpcprofile/modules>
    # @@ Don't really like having this a separate element (it makes
    # sense for <module> to be a parent of <file>, but we currently
    # don't maintain an association of modules to files/functions/lines
    set profileModulesEltObj [$xmldoc createElement modules]
    foreach module $moduleTotalList {
        $profileModulesEltObj appendFromList \
            [list module [list name [lindex $module 0]] \
                         [list [list \#text [lindex $module 1]]]]
    }
    $profileObj appendChild $profileModulesEltObj

    if { ! $havepscxxfilt } {
        ::binutils::cxxfilt::open
    }

    foreach file $fileTotalList {
        set filestring [lindex $file 0]

        # <hwpcprofilereport/hwpcprofile/file>
        set profileFileEltObj [$xmldoc createElement file]
        $profileFileEltObj setAttribute name $filestring
        $profileObj appendChild $profileFileEltObj

        set functionlist {}
        set functionnames {}
        foreach function [array names lines *:$filestring:*] {
            set functionstring [lindex [split $function :] 0]
            if { [lsearch $functionnames $functionstring] == -1 } {
                lappend functionnames $functionstring
                lappend functionlist \
                    [list $functionstring $functions($functionstring)]
            }
        }
        set functionlist [lsort -integer -decreasing -index 1 $functionlist]

        foreach function $functionlist {
            set functionstring [lindex $function 0]

            # <hwpcprofilereport/hwpcprofile/function>
            set profileFunctionEltObj [$xmldoc createElement function]
            # @@ This should use c++filt, specified as a user option
            #    [string map {& &amp; < &lt; > &gt;} $functionstring]"
            if { $havepscxxfilt } {
                $profileFunctionEltObj setAttribute name \
                    [::pscxxfilt::demangle $functionstring]
            } else {
                $profileFunctionEltObj setAttribute name \
                    [::binutils::cxxfilt::filter $functionstring]
            }
            $profileFileEltObj appendChild $profileFunctionEltObj

            set linelist {}
            foreach line [array names lines $functionstring:$filestring:*] {
                lappend linelist [list $line $lines($line)]
            }
            set linelist [lsort -integer -decreasing -index 1 $linelist]

            foreach line $linelist {
                set lineindex [lindex $line 0]
                set linestring [lindex [split $lineindex :] 2]

                # <hwpcprofilereport/hwpcprofile/line>
                set profileLineEltObj [$xmldoc createElement line]
                $profileLineEltObj setAttribute lineno $linestring
                set profileLineTextObj [$xmldoc createTextNode \
                                                $lines($lineindex)]
                $profileLineEltObj appendChild $profileLineTextObj
                $profileFunctionEltObj appendChild $profileLineEltObj
            }
        }
    }

    if { ! $havepscxxfilt } {
        ::binutils::cxxfilt::close
    }

    $xmlroot appendChild $profileObj

    return $xmldoc

}


# New implementation of generation of XML documents containing profiling
# data mapped to source code module, file, function and line.
#
# The reason why there are two procedures:
#
# - hwpcreportCreateProfileDocumentXML
# - hwpcreportCreateProfileElementXML
#
# is to distinguish between the creation of XML documents that contain a single
# profile or an XML element that contains the profile, as would be present in a
# PerfSuite <multihwpcprofilereport>.
#
# The parameter supplied to each procedure should be a document element after
# parsing by tDOM.
#
# Each procedure returns a tDOM object, either a documentElement (for
# createProfileDocument) or a domNode (for createProfileElement).

proc hwpcreportCreateProfileReportDocument { root } {
    global env

    # DTD
    if { [info exists env(PS_DTDDIR)] } {
        $domDoc systemId "$env(PS_DTDDIR)/pshwpc/hwpcprofilereport-0.2.dtd"
    }
}

proc hwpcreportCreateProfileReportElement { newroot root } {
    global options havepsbfd env tcl_platform

    if { ! [info exists havepsbfd] } {
        if { [catch {package require psbfd} res] } {
            if { [catch {package require binutils} res] } {
                error "could not find either BFD or binutils support"
            } else {
                set havepsbfd 0
            }
        } else {
            set havepsbfd 1
        }
    }

    # Set up for tracking memory usage if PS_DEBUG > 1
    if { $::perfsuite::debugLevel > 1 } {
        if { [info procs ::procfs::stat] == "" } {
            package require procfs
        }
    }

    set domDoc [$newroot ownerDocument]

    # <hwpcprofilereport>
    set reportElement [$domDoc createElement hwpcprofilereport]
    $reportElement setAttribute \
        version "0.2" generator "psprocess" \
        xmlns:perfsuite "http://perfsuite.ncsa.uiuc.edu/"

    # <hwpcprofilereport/date>
    $reportElement appendFromList \
        [list date {} \
             [list [list \#text [clock format [clock seconds] \
                                     -format "%a %b %d %H:%M:%S %Y"]]]]

    # <hwpcprofilereport/hostname>
    set hostname (null)
    foreach index {HOSTNAME hostname} {
        if [info exists env($index)] {
            set hostname $env($index)
            break
        }
    }
    $reportElement appendFromList \
        [list hostname {} [list [list \#text $hostname]]]
        
    # <hwpcprofilereport/domainname> (faked)
    $reportElement appendFromList \
        [list domainname {} [list [list \#text (null)]]]
    
    # <hwpcprofilereport/username>
    $reportElement appendFromList \
        [list username {} [list [list \#text $tcl_platform(user)]]]

    # <hwpcprofilereport/annotation>
    if { [info exists options(annotationString)] } {
        $reportElement appendFromList \
            [list annotation {} [list [list \#text $options(annotationString)]]]
    }

    # <hwpcprofilereport/executioninfo>
    set exeinfoElement [$domDoc createElement executioninfo]


    foreach elementType {date annotation hostname domainname pid thread \
                             username command wallclock cputime machineinfo } {
        set newElement [$root getElementsByTagName $elementType]
        if { $newElement != "" } {
            $exeinfoElement appendChild $newElement
        }
    }
    $reportElement appendChild $exeinfoElement

    # <hwpcprofilereport/hwpcprofile>
    set hwpcprofileElement [$domDoc createElement hwpcprofile]
    if { $havepsbfd } {
        set mapper psbfd
    } else {
        set mapper binutils
    }
    $hwpcprofileElement setAttribute mapper $mapper

    set profiledata [$root selectNodes hwpcprofiledata]
    set class [$profiledata getAttribute class "unknown"]
    set version [$profiledata getAttribute version "unknown"]
    set profileevent [$profiledata selectNodes profileinfo/profileevent]
    set domain [$profileevent getAttribute domain "unknown"]
    set type [$profileevent getAttribute type "unknown"]
    set name [$profileevent getAttribute name "unknown"]
    set period [[$profileevent selectNodes threshold] text]

    # <hwpcprofilereport/hwpcprofile/hwpcevent>
    $hwpcprofileElement appendFromList \
        [list hwpcevent [list class $class version $version type $type \
                             name $name domain $domain] {}]

    # <hwpcprofilereport/hwpcprofile/period>
    $hwpcprofileElement appendFromList \
        [list period {} [list [list \#text $period]]]

    # <hwpcprofilereport/hwpcprofile/totalsamples>
    set totalsamplesElement [$profiledata selectNodes totalsamples]
    $hwpcprofileElement appendChild $totalsamplesElement

    

    # Make sure things are set for generating the profile

    if { ! [info exists options(executable) ] } {
        set options(executable) "unknown"
    }

    if { [info procs createmappedprofile] == "" } {
        source [file join [file dirname [info script]] profile.tcl]
    }

    set proftree [createmappedprofile $profiledata $options(executable) \
                      $options(showprogress)]

    # Delete the nodes in the original document that contained the raw
    # sampling information.  These can become quite large for complex
    # and lengthy applications, and we would like to conserve the memory.

if { 1 } {
    set modulenodes [$profiledata selectNodes module]
    foreach node $modulenodes {
        $node delete
    }
}

    # <hwpcprofilereport/hwpcprofile/module> (+ file function line)
    foreach module [lsort -integer -decreasing -index 2 \
                        [lindex [lindex $proftree 0] 3]] {
        set moduleElement [$domDoc createElement module]
        $moduleElement setAttribute name [lindex $module 1]
        foreach file [lsort -integer -decreasing -index 2 [lindex $module 3]] {
            set fileElement [$domDoc createElement file]
            $fileElement setAttribute name [lindex $file 1]
            foreach function [lsort -integer -decreasing -index 2 \
                                  [lindex $file 3]] {
                set functionElement [$domDoc createElement function]
                $functionElement setAttribute name [lindex $function 1]
                foreach line [lsort -integer -decreasing -index 2 \
                                  [lindex $function 3]] {
                    set lineElement [$domDoc createElement line]
                    $lineElement setAttribute lineno [lindex $line 1]
                    set samplesElement [$domDoc createTextNode [lindex $line 2]]
                    $lineElement appendChild $samplesElement
                    $functionElement appendChild $lineElement
                }
                $fileElement appendChild $functionElement
            }
            $moduleElement appendChild $fileElement
        }
        $hwpcprofileElement appendChild $moduleElement
    }
    
    $reportElement appendChild $hwpcprofileElement


    $newroot appendChild $reportElement

    if { $::perfsuite::debugLevel > 1 } {
        array set stat [::procfs::stat]
        puts stderr "profilexml: current memory usage (MB) =\
                     [expr {$stat(vsize) / 1048576.0}]"
    }

    return
}
