#!/bin/sh
# \
exec tclsh "$0" "$@"

# $Id: psprocess.tcl,v 1.118 2009/09/30 21:12:36 rkufrin Exp $

############################################################################
#
#                    University of Illinois/NCSA
#                        Open Source License
# 
#        Copyright(C) 2002-2009, The Board of Trustees of the
#            University of Illinois. All rights reserved.
#
#                           Developed by:             
# 
#                       The PerfSuite Project
#          National Center for Supercomputing Applications
#            University of Illinois at Urbana-Champaign
# 
#                  http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

############################################################################
#
# psprocess: command-line processing for PerfSuite XML files,
#            pre- and post-collection.
#
# This file only acts as a gateway to the real processing.
# It figures out what to do based on the XML document type
# and sends the work elsewhere
#
# Author: Rick Kufrin
# Date:   12/27/2002
#
# This file is part of PerfSuite.

if [info exists env(TDOMLIBDIR)] {
    lappend auto_path $env(TDOMLIBDIR)
}

# @@ Can get rid of PS_TCLLIBDIR in the future
if [info exists env(PS_TCLLIBDIR)] {
    if { [lsearch $env(PS_TCLLIBDIR) $auto_path] == -1 } {
        lappend auto_path $env(PS_TCLLIBDIR)
    }
}

package require perfsuite
package require msgcat

namespace import perfsuite::*

namespace import -force ::msgcat::mc
msgcat::mcload [file join [file dirname [info script]] msgs]

# tDOM is required for psprocess
if { [catch {package require tdom} res] } {
    puts stderr "Error: cannot load tDOM support: \"$res\""
    exit 1
}

set ignoreErrors 1

# Performance counter support is optional
#
# @@ This data structure can be improved
#    or could instead use namespaces.
#
# @@ We could also defer loading until needed.
#    Before they were loaded early to pick up
#    msgcat definitions from papi, but this is
#    no longer necessary
set countertypes [list papi perfmon]
foreach type $countertypes {
    if { [catch {package require $type} res] } {
        set have$type 0
    } else {
        set have$type 1
    }
}
unset countertypes

# Grab the supporting procs (specific to this program)
foreach procname { combine extract filelist hwpcreport multihwpcreport options \
                   readxml resourcereport stats uri validate version xslt } {
    source [file join [file dirname [info script]]  $procname.tcl]
}

if { [catch {parseOptions} res] } {
    puts stderr "Error: $res"
    exit 1
}

psdebug 2 "Options processed, input file list: $inputFileList"

if { [info exists options(outputFile)] } {
    if { ! $options(extractMode) } {
        if { [catch {open $options(outputFile) "w"} outputfd] } {
            puts stderr "$options(outputFile): $outputfd"
            exit 1
        }
    }
} else {
    set outputfd stdout
}

if { $options(timing) } {
    foreach type {all xslt read domparse dombuild} {
        set timers(total,$type) 0
    }
    set timers(start,all) [clock clicks -milliseconds]
}

if { $options(combineMode) } {
    psdebug 1 "Doing combine of $inputFileList" stderr
    psdebug 1 "File count: [llength $inputFileList]" stderr

    if { [catch {doCombine $inputFileList $outputfd $ignoreErrors} res] } {
        puts stderr "Error combining input: $res"
        exit 1
    }
} elseif { $options(validate) } {
    psdebug 1 "Doing validation of $inputFileList" stderr
    psdebug 1 "File count: [llength $inputFileList]" stderr

    foreach file $inputFileList {
        if { [catch {validate $file} res] } {
            puts stderr $res
            exit 1
        }
    }
    exit 0
} elseif { [llength $inputFileList] > 1 } {
    psdebug 1 "Processing file list: $inputFileList" stderr
    psdebug 1 "File count: [llength $inputFileList]" stderr

    if { [catch {processFileList $inputFileList} err] } {
        puts stderr "Error: $err"
        exit 1
    }
} else {
    # Process one file, or standard input if no files supplied.
    if { [llength $inputFileList] == 1 } {
        set xmlFile [lindex $inputFileList 0]
    } else {
        set xmlFile stdin
    }
    psdebug 0 "xmlFile = $xmlFile, useXSLT = $options(useXSLT)" stderr

    # Acquire the XML data as a string
    if { [catch {readxml $xmlFile} xmlData] } {
        puts stderr "$xmlData"
        exit 1
    }

    # Use tDOM to parse into a DOM tree
    set timers(start,domparse) [clock clicks -milliseconds]
    if { $options(fast) } {
        if { [catch {dom parse -simple $xmlData xmlDoc} res] } {
            puts stderr "Error: XML parse (fast) of file \"xmlFile\" failed\
                         ($res)"
            exit 1
        }
    } else {
        if { [catch {dom parse $xmlData xmlDoc} res] } {
            puts stderr "Error: XML parse of file \"xmlFile\" failed ($res)"
            exit 1
        }
    }
    set timers(stop,domparse) [clock clicks -milliseconds]
    set timers(total,domparse) [expr {$timers(stop,domparse) - \
                                      $timers(start,domparse)}]

    # Get the XML doc itself
    set docRoot [$xmlDoc documentElement]

    # Find out what kind of document it is
    set docType [$docRoot nodeName]

    # Handle extract mode
    if { $options(extractMode) } {
        if { ! [string match multi* $docType] } {
            error "Extract mode can only be used with \"multi-\" XML documents"
        }
        if { [info exists options(outputFile)] } {
            set prefix $options(outputFile)
        } else {
            set prefix "report"
        }
        set num [extract $docRoot $prefix]
        exit 0
    }

    # Do the right thing, based on the document type
    if { $options(useXSLT) } {
        # Let tDOM's XSLT do its thing
        if { [catch {doXSLT $docType $docRoot $outputfd} res] } {
            puts stderr "$res"
            exit 1
        }
    } else {
        if { $docType == "multihwpcreport" } {
            catch {unset options(resourceReport)}
        }
        # Produce text-based output
        switch $docType {
            hwpcreport -
            multihwpcreport -
            resourcereport {
                if { [catch {$docType $docRoot $xmlFile $outputfd} res] } {
                    puts stderr "$res"
                    exit 1
                }
            }
            default {
                puts stderr "Error: unsupported XML document type \"$docType\""
                exit 1
            }
        }
    }

    if { [info exists options(outputFile)]  } {
        close $outputfd
    }
}

if { $options(timing) } {
    set timers(stop,all) [clock clicks -milliseconds]
    set timers(total,all) [expr {$timers(stop,all) - $timers(start,all)}]
    puts stderr "Total wallclock time:\
                [format {%.3f seconds} [expr {$timers(total,all)*1.0e-3}]]"
    if { [info exists timers(total,domparse)] } {
        puts stderr "Total wallclock time in XML DOM parsing:\
                    [format {%.3f seconds} \
                    [expr {$timers(total,domparse)*1.0e-3}]]"
    }
    if { [info exists timers(total,bfd)] } {
        puts stderr "Total wallclock time in profile sample mapping:\
                    [format {%.3f seconds} [expr {$timers(total,bfd)*1.0e-3}]]"
    }
}

exit 0
