# $Id: perfsuite.tcl,v 1.16 2009/02/17 16:37:57 rkufrin Exp $

############################################################################
#
#                    Illinois Open Source License
#                    University of Illinois/NCSA
#                        Open Source License
# 
#          Copyright(C) 1998-2006, The Board of Trustees of the
#                          University of Illinois.
#                           All rights reserved.
# 
#                             Developed by:             
# 
#                        The PerfSuite Project
# 
#           NCSA/University of Illinois at Urbana-Champaign
# 
#                  http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

#
# perfsuite.tcl: main PSTcl package to establish generic
#                settings for the PerfSuite Tcl library
#
#
# Author: Rick Kufrin
# Date:   1/2003

package provide perfsuite 0.6

namespace eval perfsuite {
    global env debugLevel

    ######################################################
    # Namespace variables
    ######################################################

    variable version 0.6
    variable patchLevel $version.2

    ######################################################
    # Public procedures
    ######################################################

    namespace export psdebug setDebugLevel

    ######################################################
    # Initialization
    ######################################################

    package require msgcat

    if { $tcl_version < 8.4 } {
        # Multiple message catalog settings new in Tcl 8.4. This is the mcmset
        # implementation from the Tcl source.  We force it into the toplevel
        # namespace and also we refer to "msgs" instead of "Msgs".
        #
        # msgcat::mcmset --
        #
        #	Set the translation for multiple strings in a specified locale.
        #
        # Arguments:
        #	locale		The locale to use.
        #	pairs		One or more src/dest pairs (must be even length)
        #
        # Results:
        #	Returns the number of pairs processed

        namespace eval ::msgcat {
            proc ::msgcat::mcmset {locale pairs } {
                variable msgs

                set length [llength $pairs]
                if {$length % 2} {
                    error {bad translation list: should be "mcmset locale {src dest ...}"}
                }
    
                set locale [string tolower $locale]
                set ns [uplevel 1 [list ::namespace current]]
    
                foreach {src dest} $pairs {
                    set msgs($locale,$ns,$src) $dest
                }
        
                return $length
            }
        }
    }

    if { [info exists env(PS_DEBUG)] } {
        set debugLevel $env(PS_DEBUG)
        if { $debugLevel >= 0 } {
            puts -nonewline "PerfSuite Tcl debugging on: level = "
            switch -exact -- $debugLevel {
                0       {puts "FATAL"}
                1       {puts "WARNING"}
                2       {puts "INFO"}
                default {puts "VERBOSE"}
            }
        }
    } else {
        set debugLevel -1
    }
}

# Conventions for PerfSuite Tcl debugging similar to libperfsuite:
#
# PS_FATAL   = 0
# PS_WARNING = 1
# PS_INFO    = 2
# PS_VERBOSE = 3
#
# The lower the level supplied to psdebug, the less verbose (and more
# severe) the debugging statement.  Most informational messages should
# be "psdebug 2 msg".

if { $::perfsuite::debugLevel >= 0 } {
    proc ::perfsuite::psdebug {level msg {fd stdout}} {
        global ::perfsuite::debugLevel

        if { $level <= $::perfsuite::debugLevel} {
            puts $fd $msg
        }
    }
} else {
    proc ::perfsuite::psdebug {args} {}
}

# debug levels less than 0 remove debugging procedures from
# the interpreter.

proc ::perfsuite::setDebugLevel {level} {
    global ::perfsuite::debugLevel

    if { $level != $::perfsuite::debugLevel } {
        set ::perfsuite::debugLevel $level
    }

    if { $level >= 0 } {
        proc ::perfsuite::psdebug {level msg {fd stdout}} {
            global ::perfsuite::debugLevel

            if { $level >= $::perfsuite::debugLevel} {
                puts $fd $msg
            }
        }
    } else {
        proc ::perfsuite::psdebug {args} {}
    }
}
