# $Id: metrics.tcl,v 1.19 2006/05/26 19:18:34 rkufrin Exp $

############################################################################
#
#                    University of Illinois/NCSA
#                        Open Source License
# 
#          Copyright(C) 2003-2006, The Board of Trustees of the
#              University of Illinois. All rights reserved.
#
#                            Developed by:             
# 
#                        The PerfSuite Project
#            National Center for Supercomputing Applications
#              University of Illinois at Urbana-Champaign
# 
#                   http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

#
# Author: Rick Kufrin
# Date:   July 2003
#
# This file is part of PerfSuite.

package provide pshwpcmetrics 0.1

# Subset of MathML content elements accepted:

# <ci> - used for performance events and keywords "cpusecs", "wallsecs",
#        "mhz", "lNcacheline"
# <cn> - numeric constants hard-coded in to the expression
# <csymbol> - possibly for peak FP performance (?)
# <apply>
# <divide>
# <minus>
# <plus>
# <times>
# <abs>

if { [catch {package require tdom} res] } {
    error $res
}

if { [catch {package require msgcat} res] } {
    error $res
}


# @@ This does not handle non-element nodes well
namespace eval psmetrics {
    namespace export load derivedMetrics

    ###################################################################
    #  Public procedures
    ###################################################################

    proc load {file} {
        if { [catch {open $file "r"} fd] } {
            error $fd
        }

        set xmldata [read $fd]

        close $fd

        if { [catch {dom parse $xmldata xmldoc} res] } {
            error "XML parsing error: $res"
        }

        $xmldoc documentElement xmlroot

        set doctype [$xmlroot nodeName]

        switch $doctype {
            psmetrics {
                 ${doctype}::processXMLDefinition $xmlroot
            }
            default {
                error "$file is not an XML document with root element\
                       <psmetrics> (found <\"$doctype\">)"
            }    
        }
    }

    proc derivedMetrics {option args} {
        global env
        variable procs

        set eventList {}
        set wallclock ""
        set clockspeed ""
        set cacheline {}

        switch -- $option {
            generate {
                set argc [llength $args]
                for {set arg 0} {$arg < $argc} {incr arg} {
                    switch -- [lindex $args $arg] {
                        -cacheline {
                            incr arg
                            set cacheline [lindex $args $arg]
                        }
                        -clockspeed {
                            incr arg
                            set clockspeed [lindex $args $arg]
                        }
                        -events {
                            incr arg
                            set eventList [lindex $args $arg]
                        }
                        -wallclock {
                            incr arg
                            set wallclock [lindex $args $arg]
                        }
                        -cyclecounter {
                            incr arg
                            set cyclecounter [lindex $args $arg]
                        }
                        -metricfile {
                            incr arg
                            set metricfile [lindex $args $arg]
                        }
                        default {
                            error "Unknown option\
                                    \"[lindex $args $arg]\": must be\
                                    -cacheline, -clockspeed, -cyclecounter,\
                                    -events, -metricfile, -wallclock"
                        }
                    }
                }
                if {[llength $eventList] == 0} {
                    error "no events supplied\
                            (-events {name1 value1 name2 value2 ...})"
                }

                array set l $eventList

                # Guard against overflow by converting to floating point
                # @@ Can also remove all the casts to double after this
                foreach event [array names l] {
                    scan $l($event) "%f" l($event)
                }
                if { [info exists l($cyclecounter)] } {
                    set p(cpusecs) [expr {double($l($cyclecounter)) / \
                            ($clockspeed * 1.0e6)}]
                }
                if { $wallclock != "" } {
                    set p(wallsecs) $wallclock
                }
                if { $clockspeed != "" } {
                    set p(mhz) $clockspeed
                }
                for {set level 0} \
                    {$level < [llength $cacheline]} \
                    {incr level} {
                    set cacheindex [expr {$level + 1}]
                    set p(l${cacheindex}cacheline) [lindex $cacheline $level]
                }
                # Yikes!
                if { ! [info exists procs] } {
                    # puts stderr "Loading PerfSuite standard derived metric definitions"
                    # @@ Temporary fix.  The "standard" location
                    # of the PerfSuite XML directory has to be
                    # set in some way.
                    if { [info exists env(PS_HWPC_METRICS_DIR)] } {
                        set metricfile [file join $env(PS_HWPC_METRICS_DIR) \
                                        $metricfile]
                    } elseif { [info exists env(PS_XMLDIR)] } {
                        set metricfile [file join $env(PS_XMLDIR) \
                                        pshwpc $metricfile]
                    } else {
                        set metricfile \
                            [file join \
                                [file dir [file dir [file dir [info script]]]] \
                                xml pshwpc $metricfile]
                    }
                    if { ! [file exists $metricfile] } {
                        error "Cannot locate default metric file:\
                               ${metricfile} "
                    }
                    set procs [psmetrics::load $metricfile]

                    if { [info exists env(PS_HWPC_METRICS)] } {
                        # puts stderr "Loading user-definited derived metric definitions"
                        if { [catch {psmetrics::load \
                                     $env(PS_HWPC_METRICS)} userprocs] } {
                            puts stderr "Warning (PS_HWPC_METRICS): $env(PS_HWPC_METRICS): $userprocs"
                            puts stderr "Ignoring user metric file...\n"
                        } else {
                            set procs [concat $procs $userprocs]
                        }
                    }
                }
                set metrics {}
                foreach proc $procs {
                    if { [catch {$proc l p} res] } {
                        continue
                    } elseif { $res == "" } {
                        continue
                    } else {
                        lappend metrics $res
                    }
                }

                return $metrics
            }
            help {
                puts "Usage: ::psmetrics::derivedMetrics option ?args?"
                puts "\toption: \"generate\" or \"help\""
                puts "\targs:   -cacheline {l1 ?l2? ...} (in bytes)"
                puts "\t        -clockspeed speed (in MHz)"
                puts "\t        -wallclock time (in seconds)"
                puts "\t        -cyclecounter eventname"
                puts "\t        -metricfile filename"
                puts "\t        -events {name1 value1 name1 value2 ...}"
            }
            default {
                error "option must be \"generate\" or \"help\""
            }
        }
    }

    ###################################################################
    #  Private procedures
    ###################################################################

    proc processXMLDefinition {root} {

        set proclist {}
        set debug 0
        
        # Get rid of all comments, processing nodes, cdata
        foreach node [$root selectNodes {descendant::node()}] {
            if { ! ([string match [$node nodeType] "ELEMENT_NODE"] || \
                    [string match [$node nodeType] "TEXT_NODE"]) } {
                $node delete
            }
        }
        
        for { set metricelement [$root firstChild] } \
            { $metricelement != "" } \
            { set metricelement [$metricelement nextSibling] } {

            if { ! [string match [$metricelement nodeName] metric] } {
                invalidElement metric [$metricelement nodeName]
            }

            if { [catch {$metricelement getAttribute namespace} namespace] } {
                error "missing required \"namespace\" attribute for <metric>"
            }
            set type [$metricelement getAttribute type]
            
            set nameelement [$metricelement firstChild]
            
            if { ! [string match [$nameelement nodeName] name] } {
                invalidElement name [$nameelement nodeName]
            }
            set label [$nameelement text]
            if { [regexp {[ \t\n]} $label] } {
                error "<name> element \"$label\" in metric definition\
                       may not contain whitespace"
            }
            
            set descelement [$nameelement nextSibling]

            catch {unset desclist}
            set gotdesc 0
            for { set descelement [$nameelement nextSibling] } \
                { [string match [$descelement nodeName] description] } \
                { set descelement [$descelement nextSibling] } {
                set gotdesc 1
                set desc [$descelement text]
                set lang [$descelement getAttribute lang]
                if { $lang == "" } {
                    set lang en
                }
                lappend desclist [list $lang $desc]
            }

            if { ! $gotdesc} {
                invalidElement description [$descelement nodeName]
            }
            
            set procname "[namespace current]::${namespace}::${label}"
            if { [lsearch $proclist $procname] != -1 } {
                puts stderr "\nWarning: multiple definition for metric:"
                puts stderr "\t\"[namespace tail $procname]\""
                puts stderr "(found in the same metric definition document)\n"
            } elseif { [info procs $procname] != "" } {
                puts stderr "\nWarning: multiple definition for metric:"
                puts stderr "\t\"[namespace tail $procname]\""
                puts stderr "(found in multiple metric definition documents)\n"
            } else {
                lappend proclist $procname
            }
            
            
            set result "proc $procname {arr params} {\n"
            append result "    upvar \$arr a\n    upvar \$params p\n\n"
            
            if { $debug } {
                puts "We have a $namespace $type definition for $label"
                puts "Available descriptions \[language\]:"
                foreach desc $desclist {
                    puts "\t[lindex $desc 1] \[[lindex $desc 0]\]"
                }
            }
            
            set defnelement $descelement
            if { ! [string match [$defnelement nodeName] definition] } {
                invalidElement definition [$defnelement nodeName]
            }
            
            set applyelement [$defnelement firstChild]
            if { ! [string match [$applyelement nodeName] apply] } {
                invalidElement apply [$applyelement nodeName]
            }
            
            set topexpelement [$applyelement firstChild]
            
            set arglist {}
            set plist {}
            set a {}
            set p {}
            set exp {}
            [namespace current]::genexpr $topexpelement a p arglist plist exp
            if { [llength $arglist] > 0 } {
                append result "    foreach measurement \[list $arglist\] \{\n"
                append result "        if \{ ! \[info exists a(\$measurement)\] \} \{\n"
                append result "            return \"\"\n"
                append result "        \}\n"
                append result "    \}\n"
            }
            if { [llength $plist] > 0 } {
                append result "    foreach param \[list $plist\] \{\n"
                append result "        if \{ ! \[info exists p(\$param)\] \} \{\n"
                append result "            return \"\"\n"
                append result "        \}\n"
                append result "    \}\n\n"
            }
            append result "    return \[list \[::msgcat::mc \"[lindex [lindex $desclist 0] 1]\"\] \[expr \{"
            append result "$exp\}\]\]\n}"
            
            if { $debug } {
                puts "result:\n$result"
            }
            
            uplevel #0 namespace eval [namespace current]::$namespace [list $result]
            # Install the message catalog entries.
            # @@ This assumes the first one is the one that is
            # the "base" entry - this should be changed.
            set basedesc [lindex [lindex $desclist 0] 1]
            foreach desc $desclist {
                # $puts "[namespace current] ::msgcat::mcset [lindex $desc 0] \"$basedesc\" \"[lindex $desc 1]\""
                ::msgcat::mcset [lindex $desc 0] "$basedesc" "[lindex $desc 1]"
            }
        }
        
        return $proclist
    }
    
    # Generate a subexpression given a node, for example:
    #     (double($a(PAPI_FP_INS)) / double($a(PAPI_TOT_CYC)))
    #
    # This has to handle <ci> and <cn> elements as well as
    # subexpressions, which are constructed recursively.
    
    proc genexpr {node arr params arglist plist res} {
        upvar $arr a
        upvar $params p
        upvar $arglist al
        upvar $plist pl
        upvar $res r
        
        array set op {divide "/" minus "-" plus "+" times "*"}
        
        set top [$node nodeName]
        set arglist [$node selectNodes {./*}]
        switch -- $top {
            divide -
            minus -
            plus -
            times {
                if { [llength $arglist] != 2 } {
                    error "<$top> element must have exactly two children"
                }
                append r "("
                genexpr [lindex $arglist 0] a p al pl r
                append r $op($top)
                genexpr [lindex $arglist 1] a p al pl r
                append r ")"
            }
            abs {
                if { [llength $arglist] != 1 } {
                    error "<$top> element must have exactly one child"
                }
                append r "(abs("
                genexpr [lindex arglist 0] a p al pl r
                append r "))"
            }
            ci {
                set val [$node text]
                switch -regexp $val {
                    {(wallsecs|cpusecs|mhz|l[0-9]cacheline)} {
                        append r "double(\$p($val))"
                        if { [lsearch $pl $val] == -1 } {
                            lappend pl $val
                        }
                    }
                    default {
                        append r "double(\$a($val))"
                        if { [lsearch $al $val] == -1 } {
                            lappend al $val
                        }
                    }
                }
            }
            cn {
                set val [$node text]
                append r "double($val)"
            }
            default {
                error "Unrecognized element: $top"
            }
        }
    }

    proc invalidElement {expected got} {
        error "Improper XML document (expected element <$expected>,\
               found <$got>)"
    }
}
