# $Id: papi.tcl,v 1.37 2006/01/16 21:21:37 rkufrin Exp $

############################################################################
#                    University of Illinois/NCSA
#                        Open Source License
# 
#        Copyright(C) 2002-2006, The Board of Trustees of the
#            University of Illinois. All rights reserved.
#
#                           Developed by:             
# 
#                      The PerfSuite Project
#        National Center for Supercomputing Applications 
#           University of Illinois at Urbana-Champaign
# 
#                http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

############################################################################
#
# papi - very basic support for PAPI inquiry and utility functions
#        (no counting)
#
# Current version provides:
#
# - namespace variable ::papi::installed that returns 1 if PAPI
#     is available, 0 if not
# - namespace variable ::papi::stdEvents
#     This array is indexed by the name of the PAPI standard event
#     Each element is a list:
#       {code description label status isderived {eventcodes}}
#
#     "eventcodes" is also the "note" associated with the event
#
# Routines that are available regardless of presence of PAPI itself:
#
# - procedure "stdEvents" that returns a list of all PAPI events,
#       optionally filtered by a glob-style pattern match
# - procedure "getDescription" that takes an argument of an
#       event name and returns a string containing the description
#       if the event exists, generates an error otherwise
# - procedure "availEvents" that returns a list of all available
#       PAPI events
# - procedure "isAvail" that takes an argument that is an event
#       name and returns 1 if event is available, 0 if not.
#       generates an error if the event is not a standard event
#
# Routines that require PAPI to be installed on the machine:
#
# - procedure "isDerived" that takes an argument that of an event
#       name and returns 1 if it's a derived event, 0 if not, and
#       -1 if the event isn't available.  Error if a non-existent
#       event is given
# - procedure "getCodes" that takes an argument of an event name
#       and returns a list of event codes associated with the name
#       if the event exists, generates an error otherwise
#
# Additional procedures are defined directly in the PAPI C extension,
# and are documented in those source files.
#
# Author: Rick Kufrin
#
# This file is part of PerfSuite

package provide papi 0.1; # Make sure to synch with pkgIndex.tcl everywhere

package require perfsuite
package require pshwpcmetrics
package require msgcat
namespace import -force ::msgcat::mc

if { [catch {package require pspapi} ] } {
    namespace eval papi { set ::papi::installed 0 }
    source [file join [file dirname [info script]] papiStdEvents.tcl]
} else {
    namespace eval papi { set ::papi::installed 1 }
}

namespace eval papi {

    #################################################################
    # Public variables
    #################################################################

    variable stdEvents
  
    #################################################################
    # Private variables
    #################################################################

    variable metricfile "PAPI_metrics.xml"
    variable cyclecounter "PAPI_TOT_CYC"
    variable index
    array set index [list CODE 0 DESCRIPTION 1 LABEL 2 STATUS 3 \
                          ISDERIVED 4 EVENTCODES 5]

    #################################################################
    # Public procedures
    #################################################################

    namespace export availEvents getCodes getDescription isAvail \
            isDerived stdEvents derivedMetrics

    # Initialize the message catalog
    msgcat::mcload [file join [file dirname [info script]] msgs]
}

namespace eval papi {

    #################################################################
    # Public procedures
    #################################################################

    proc isAvail {event} {
        variable stdEvents
        variable index

        if { ! [papi::exists $event] } {
            error "event $event does not exist"
        }
        if { [lindex $stdEvents($event) $index(STATUS)] \
              == "available" } {
            return 1
        }
        return 0
    }

    proc availEvents {{pat ""}} {
        variable stdEvents

        set avail {}
        if { $pat == "" } {
            set events [array names stdEvents]
        } else {
            set events [array names stdEvents $pat]
        }
        foreach event $events {
            if { [papi::isAvail $event] } {
                lappend avail $event
            }
        }
        return $avail
    }

    proc isDerived {event} {
        variable stdEvents
        variable index

        if { ! [papi::exists $event] } {
            error "event $event does not exist"
        }
        if { ! [papi::isAvail $event] } {
            return -1
        }
        if { [lindex $stdEvents($event) $index(ISDERIVED)] \
             == "derived" } {
            return 1
        }
        return 0
    }

    proc stdEvents {args} {
        variable stdEvents

        if { [llength $args] > 0 } {
            if { [string match [lindex $args 0] "-asXML"] } {
                # name, label, description, short description
                append xmlres "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n"
                append xmlres "<PAPIstdevents>\n"
                foreach event [lsort [array names stdEvents]] {
                   append xmlres "\t<PAPIevent name=\"$event\">\n"
                   append xmlres "\t\t<description>[getDescription $event]" 
                   append xmlres "</description>\n" 
                   append xmlres "\t</PAPIevent>\n"
                }
                append xmlres "</PAPIstdevents>"
                return $xmlres
            } else {
                return [array names stdEvents [lindex $args 0]]
            }
        } else {
            return [array names stdEvents]
        }
    }

    proc getCodes {event} {
        variable stdEvents
        variable index

        if { ! [papi::exists $event] } {
            error "event $event does not exist"
        }
        return [lindex $stdEvents($event) $index(EVENTCODES)]
    }

    proc getDescription {event} {
        variable stdEvents
        variable index

        if { ! [papi::exists $event] } {
            error "event $event does not exist"
        }
        # See if there's a message catalog entry
        set description [mc $event]
        if { ! [string match $description $event] } {
            return $description
        }
        return [lindex $stdEvents($event) $index(DESCRIPTION)]
    }

    proc derivedMetrics {option args} {
        variable metricfile
        variable cyclecounter

        lappend args -metricfile $metricfile -cyclecounter $cyclecounter

        return [eval ::psmetrics::derivedMetrics generate $args]
    }

    #################################################################
    # Private procedures
    #################################################################

    proc exists {event} {
        variable stdEvents

        return [info exists stdEvents($event)]
    }
}
