# $Id: perfmon.tcl,v 1.14 2006/01/16 21:22:26 rkufrin Exp $

############################################################################
#                    University of Illinois/NCSA
#                        Open Source License
# 
#        Copyright(C) 2004-2006, The Board of Trustees of the
#            University of Illinois. All rights reserved.
#
#                           Developed by:             
# 
#                      The PerfSuite Project
#        National Center for Supercomputing Applications 
#           University of Illinois at Urbana-Champaign
# 
#                http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

############################################################################
#
# perfmon - very basic support for Perfmon inquiry and utility functions
#           (no counting)
#
# Current version provides:
#
# - namespace variable ::perfmon::installed that returns 1 if Perfmon
#     is available, 0 if not
# - namespace variable ::perfmon::stdEvents
#     This array is indexed by the name of the IA-64 standard event
#     Each element is a list:
#       {code description {counters}}
#     where {counters} is a 4-element list as returned by libpfm
#
#
# Routines that are available regardless of presence of Perfmon itself:
#
# - procedure "stdEvents" that returns a list of all IA-64 events,
#       optionally filtered by a glob-style pattern match
# - procedure "getDescription" that takes an argument of an
#       event name and returns a string containing the description
#       if the event exists, generates an error otherwise
# - procedure "availEvents" that returns a list of all available
#       IA-64 events
# - procedure "isAvail" that takes an argument that is an event
#       name and returns 1 if event is available, 0 if not.
#       generates an error if the event is not a standard event
#
# Routines that require Perfmon to be installed on the machine:
#
# - procedure "getCodes" that takes an argument of an event name
#       and returns a list of event codes associated with the name
#       if the event exists, generates an error otherwise
#
# Additional procedures are defined directly in the Perfmon C extension,
# and are documented in those source files.
#
# Author: Rick Kufrin
#
# This file is part of PerfSuite.

package provide perfmon 0.1

package require perfsuite
if { [catch {package require pshwpcmetrics} res] } {
    error $res
}

package require msgcat
namespace import -force ::msgcat::mc

if { [catch {package require psperfmon} ] } {
    namespace eval perfmon { set ::perfmon::installed 0 }
    source [file join [file dirname [info script]] perfmon_events.tcl]
} else {
    namespace eval perfmon { set ::perfmon::installed 1 }
}

namespace eval perfmon {

    #################################################################
    # Public variables
    #################################################################

    variable stdEvents
  
    #################################################################
    # Private variables
    #################################################################

    variable metricfile "perfmon_metrics.xml"
    variable cyclecounter "CPU_CYCLES"
    variable index
    array set index [list CODE 0 DESCRIPTION 1 COUNTERS 2] \

    #################################################################
    # Public procedures
    #################################################################

    namespace export availEvents getCodes getCounters getDescription \
                     isAvail isDerived stdEvents

    # Initialize the message catalog
    msgcat::mcload [file join [file dirname [info script]] msgs]
}

namespace eval perfmon {

    #################################################################
    # Public procedures
    #################################################################

    proc isAvail {event} {
        if { ! [perfmon::exists $event] } {
            error "event $event does not exist"
        }
        if { ! $::perfmon::installed } {
            return 0
        }
        return 1
    }

    proc availEvents {{pat ""}} {
        variable stdEvents

        set avail {}
        if { $::perfmon::installed } {
            if { $pat == "" } {
                set events [array names stdEvents]
            } else {
                set events [array names stdEvents $pat]
            }
            foreach event $events {
                if { [perfmon::isAvail $event] } {
                    lappend avail $event
                }
            }
        }
        return $avail
    }

    proc stdEvents {args} {
        variable stdEvents

        if { [llength $args] > 0 } {
            if { [string match [lindex $args 0] "-asXML"] } {
                # name, label, description, short description
                append xmlres "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n"
                append xmlres "<Perfmonstdevents>\n"
                foreach event [lsort [array names stdEvents]] {
                   append xmlres "\t<Perfmonevent name=\"$event\">\n"
                   append xmlres "\t\t<description>[getDescription $event]" 
                   append xmlres "</description>\n" 
                   append xmlres "\t</Perfmonevent>\n"
                }
                append xmlres "</Perfmonstdevents>"
                return $xmlres
            } else {
                return [array names stdEvents [lindex $args 0]]
            }
        } else {
            return [array names stdEvents]
        }
    }

    proc getCodes {event} {
        variable stdEvents
        variable index

        if { ! [perfmon::exists $event] } {
            error "event $event does not exist"
        }
        return [lindex $stdEvents($event) $index(CODE)]
    }

    proc getCounters {event} {
        variable stdEvents
        variable index

        if { ! [perfmon::exists $event] } {
            error "event $event does not exist"
        }
        return [lindex $stdEvents($event) \
                $index(COUNTERS)]
    }

    proc getDescription {event} {
        variable stdEvents
        variable index

        if { ! [perfmon::exists $event] } {
            error "event $event does not exist"
        }
        # See if there's a message catalog entry
        set description [mc $event]
        if { ! [string match $description $event] } {
            return $description
        }
        return [lindex $stdEvents($event) $index(DESCRIPTION)]
    }

    proc derivedMetrics {option args} {
        variable metricfile
        variable cyclecounter

        lappend args -metricfile $metricfile -cyclecounter $cyclecounter

        return [eval ::psmetrics::derivedMetrics generate $args]
    }

    #################################################################
    # Private procedures
    #################################################################

    proc exists {event} {
        variable stdEvents

        return [info exists stdEvents($event)]
    }
}
