# $Id: hwpcreport.tcl,v 1.11 2006/01/16 21:25:20 rkufrin Exp $

############################################################################
#
#                      University of Illinois/NCSA
#                          Open Source License
# 
#          Copyright(C) 2004-2006, The Board of Trustees of the
#              University of Illinois. All rights reserved.
#
#                             Developed by:             
# 
#                        The PerfSuite Project
#            National Center for Supercomputing Applications
#              University of Illinois at Urbana-Champaign
# 
#                   http://perfsuite.ncsa.uiuc.edu/
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal with the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
# 
# + Redistributions of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimers.
# + Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimers in
#   the documentation and/or other materials provided with the distribution.
# + Neither the names of The PerfSuite Project, NCSA/University of Illinois
#   at Urbana-Champaign, nor the names of its contributors may be used to
#   endorse or promote products derived from this Software without specific
#   prior written permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS WITH THE SOFTWARE.
############################################################################

#
# Parsing routines for PerfSuite <hwpcreport> XML documents.
#
# Input:  xmlroot - the document element as parsed by tDOM.
# Output: outarray - a Tcl array populated with the data contained
#                    in the XML document, with the following
#                    array elements:
#
#         reporttype:    "hwpcreport"
#         annotation:    contents of the <annotation> element
#                        @@ should this *always* be set?
#         command:       contents of the <command> element
#         date:          contents of the <date> element
#         domainname:    contents of the <domain> element
#         hostname:      contents of the <hostname> element
#         username:      contents of the <username> element
#         wallclock:     wallclock time of the run, in seconds
#         machineinfo:   a list with the following structure:
#
#                        {cpucount N memoryinfo MB syspagesize KB
#                         cpuinfo CPUINFO}
#
#                        where CPUINFO is also a list with the following
#                        structure:
#
#                        {vendor VENDOR family FAMILY revision REVISION
#                         clockspeed MHz cacheinfo CACHEINFO tlbinfo TLBINFO}
#
#                        in turn, CACHEINFO is a list with the following
#                        structure.  The intent is to group each cache level
#                        as a sublist:
#
#                        {
#                         {{type TYPE size KB linesize B associativity N}
#                          {type TYPE size KB linesize B associativity N}
#                          ...
#                         }
#                         {{type TYPE size KB linesize B associativity N}
#                          {type TYPE size KB linesize B associativity N}
#                          ...
#                         }
#                        }
#
#                        TLBINFO is similar, except the sublists are:
#
#                        {
#                         {{type TYPE entries N associativity N}
#                          {type TYPE entried N associativity N}
#                          ...
#                         }
#                        }
#         hwpceventdata: a list with the following structure:
#
#                        {class CLASS domain DOMAIN multiplexed (yes|no)
#                         hwpcevents HWPCEVENTDATA hwpceventinfo HWPCEVENTINFO}
#
#                        where HWPCEVENTDATA is a list with the
#                        following structure:
#
#                        {name1 count1 name2 count2 .... nameN countN}
#
#                        and HWPCEVENTINFO is a list with the following
#                        structure:
#
#                        {name1 TYPE DERIVED}
#
#                        where TYPE is either "preset" or "native" and
#                        DERIVED is either "yes" or "no" (note: yes/no
#                        are valid booleans that can be used in Tcl
#                        expressions) 
# 
#         hwpcprofiledata: a list with the following structure:
#
#                        {class CLASS domain DOMAIN threshold THRESHOLD
#                         totalsamples TOTALSAMPLE hwpcsamples HWPCSAMPLEDATA 
#                         hwpceventinfo HWPCEVENTINFO}
#
#                       where HWPCSAMPLEDATA is a list with the 
#                       following structure:
#
#                       {{module MODULENAME offset OFFSET base BASE 
#                         samples SAMPLES}
#                        {module MODULENAME offset OFFSET base BASE
#                         samples SAMPLES}
#                        ... }
#
#                       where SAMPLES is a list with the following structure:
#
#                       {pc1 count1 pc2 count2 ... pcN countN}
#
#                       and HWPCEVENTINFO is as above.
#
#
# Author: Rick Kufrin
# Date:   August, 2004
#
# This file is part of PerfSuite.

package require tdom

namespace eval perfsuite::xml::hwpcreport {
    namespace export parse
}

proc perfsuite::xml::hwpcreport::parse {docelt outarray} {
    upvar $outarray out

    catch {array unset out}

    foreach attribute { version generator mode } {
        set out($attribute) [$docelt getAttribute $attribute]
    }

    for { set child [$docelt firstChild] } \
        { $child != "" } { set child [$child nextSibling] } {
        set nodeType [$child nodeName]
        switch $nodeType {
            annotation -
            command -
            date -
            domainname -
            hostname -
            username -
            wallclock {
                set out($nodeType) [$child text]
            }
            hwpceventdata -
            hwpcprofiledata -
            machineinfo {
                parse_$nodeType $child out
            }
            default {
                #set out($nodeType) "(no value)"
            }
        }
    }
}

proc perfsuite::xml::hwpcreport::parse_machineinfo {node outarray} {
    upvar $outarray out

    for { set child [$node firstChild] } \
        { $child != "" } { set child [$child nextSibling] } {
        set nodeType [$child nodeName]
        switch $nodeType {
            cpucount -
            memoryinfo -
            syspagesize {
                lappend syslist $nodeType [$child text]
            }
            cpuinfo {
                set cpuinfo [parse_cpuinfo $child]
                lappend syslist $nodeType $cpuinfo
            }
            default {
                error "Unrecognized <machineinfo> subelement: $nodeType"
            }
        }
    }

    set out(machineinfo) $syslist
}

proc perfsuite::xml::hwpcreport::parse_hwpceventdata {node outarray} {
    upvar $outarray out

    foreach attribute [$node attributes] {
        lappend attlist $attribute [$node getAttribute $attribute]
    }

    set eventNodes [$node getElementsByTagName hwpcevent]

    foreach event $eventNodes {
        set name [$event getAttribute name]
        lappend eventcounts $name [$event text]
        lappend eventinfo   $name \
            [list [$event getAttribute type] \
                  [$event getAttribute derived]]
    }

    set result [concat $attlist hwpcevents [list $eventcounts] \
                       hwpceventinfo [list $eventinfo]]
    set out(hwpceventdata) $result
}

proc perfsuite::xml::hwpcreport::parse_hwpcprofiledata {node outarray} {
    upvar $outarray out

    foreach attribute [$node attributes] {
        lappend attlist $attribute [$node getAttribute $attribute]
    }

    set profileinfo [$node getElementsByTagName profileevent]

    lappend info domain [$profileinfo getAttribute domain]
    set threshold [$node getElementsByTagName threshold]
    lappend info threshold [$threshold text]
    set totalsamples [$node getElementsByTagName totalsamples]
    lappend info totalsamples [$totalsamples text]

    foreach att { name type derived } {
        lappend eventinfo [$profileinfo getAttribute $att]
    }

    set modulenodes [$node getElementsByTagName module]
    foreach mnode $modulenodes {
        set sampleelt [$mnode getElementsByTagName samples]
        set samples   [$sampleelt getElementsByTagName sample]
        set samplelist {}
        foreach sample $samples {
            lappend samplelist [list [$sample getAttribute pc] \
                                     [$sample text]]
        }
        lappend modulelist [list module [$mnode getAttribute file] \
                                 offset [$mnode getAttribute offset] \
                                 base   [$sampleelt getAttribute pcbase] \
                                 samples $samplelist]
    }

    set result [concat $attlist $info hwpceventinfo [list $eventinfo] \
                       hwpcsamples [list $modulelist]]
    set out(hwpcprofiledata) $result
}

proc perfsuite::xml::hwpcreport::parse_cpuinfo {node} {
    for { set child [$node firstChild] } \
        { $child != "" } { set child [$child nextSibling] } {
        set nodeType [$child nodeName]
        switch $nodeType {
            brand -
            clockspeed -
            family -
            revision -
            vendor {
                lappend res $nodeType [$child text]
            }
            cacheinfo {
                set cachelist [parse_cacheinfo $child]
                lappend res $nodeType $cachelist
            }
            tlbinfo {
                set tlblist [parse_tlbinfo $child]
                lappend res $nodeType $tlblist
            }
            default {
                error "Unrecognized <cpuinfo> subelement: $nodeType"
            }
        }
    }
    #puts "parse_cpuinfo returning $res"
    return $res
}

proc perfsuite::xml::hwpcreport::parse_cacheinfo {node} {
    set cachelist {}
    for { set child [$node firstChild] } \
        { $child != "" } { set child [$child nextSibling] } {
        set nodeType [$child nodeName]
        switch $nodeType {
            levels {
               set levels [$child text]
               for {set i 0} {$i < $levels} {incr i} {
                   lappend cachelist {}
                   #puts "inserting level $i cache, length is [llength $cachelist]"
               }
            }
            cache {
               add_cache cachelist $child
            }
            default {
               error "Unrecognized <cacheinfo> subelement: $nodeType"
            }
        }
    }
    return $cachelist
}

proc perfsuite::xml::hwpcreport::add_cache {cachelist node} {
    upvar $cachelist cl

    set level [$node getAttribute level]
    set index [expr {$level - 1}]
    set sublist [lindex $cl $index]
    set type  [$node getAttribute type]
    lappend stats type $type
    foreach stat {size linesize associativity} {
        set elt [$node getElementsByTagName $stat]
        if { $elt == "" } {
            # Instruction trace caches have no linesize
            if { $stat == "linesize" } {
                continue
            }
            error "Missing <$elt> element in cache description"
        }
        set val [$elt text]
        lappend stats $stat $val
    }
    set sublist [linsert $sublist 0 $stats]
    set cl [lreplace $cl $index $index $sublist]
    #puts "Replaced $sublist to cachelist index $index (level $level)"
}

# Note that <tlbinfo> elements do not currently contain a <levels>
# child element
proc perfsuite::xml::hwpcreport::parse_tlbinfo {node} {
    set tlbelts [$node getElementsByTagName tlb]
    if { $tlbelts == "" } {
        error "No <tlb> elements found"
    }
    set maxlevel 0
    foreach elt $tlbelts {
        set level [$elt getAttribute level]
        if { $level > $maxlevel } {
            set maxlevel $level
        }
    }

    set tlblist {}
    for {set i 0} {$i < $maxlevel} {incr i} {
        lappend tlblist {}
    }

    for { set child [$node firstChild] } \
        { $child != "" } { set child [$child nextSibling] } {
        set nodeType [$child nodeName]
        switch $nodeType {
            tlb {
               add_tlb tlblist $child
            }
            default {
               error "Unrecognized <tlbinfo> subelement: $nodeType"
            }
        }
    }
    #puts "parse_tlblist returning $tlblist"
    return $tlblist
}

proc perfsuite::xml::hwpcreport::add_tlb {tlblist node} {
    upvar $tlblist tl

    set level [$node getAttribute level]
    set index [expr {$level - 1}]
    set sublist [lindex $tl $index]
    set type  [$node getAttribute type]
    lappend stats type $type
    foreach stat {entries associativity} {
        set elt [$node getElementsByTagName $stat]
        if { $elt == "" } {
            error "Missing <$elt> element in TLB description"
        }
        set val [$elt text]
        lappend stats $stat $val
    }
    set sublist [linsert $sublist 0 $stats]
    set tl [lreplace $tl $index $index $sublist]
    #puts "Replaced $sublist to tlblist index $index (level $level)"
}

